"""
This file contains a library of functions related to fitting and MCMC

"""

import os
import numpy as np
import emcee


#==================================================
# Define the starting point of MCMC with emcee
#==================================================

def emcee_starting_point(guess_value, guess_error, par_min, par_max, nwalkers):
    """
    Sample the parameter space for emcee MCMC starting point
    from a uniform distribution in the parameter space.
        
    Parameters
    ----------
    - guess_value (Nparam array): guess value of the parameters
    - guess_error (float): flat dispersion allowed within this range for unborned parameters
    - parmin (Nparam array): min value of the parameters (i.e. sharp prior)
    - parmax (Nparam array): max value of the parameters (i.e. sharp prior)
    - nwalkers (int): number of walkers

    Output
    ------
    start: the starting point of the chains in the parameter space

    """

    ndim = len(guess_value)

    # First range using guess + dispersion
    vmin = guess_value - guess_error 
    vmax = guess_value + guess_error

    # Check that the parameters are in the prior range
    wup = vmin < np.array(par_min)
    wlo = vmax > np.array(par_max)
    vmin[wup] = np.array(par_min)[wup]
    vmax[wlo] = np.array(par_max)[wlo]

    # Get parameters
    start = [np.random.uniform(low=vmin, high=vmax) for i in range(nwalkers)]
    
    return start


#==================================================
# Check if the sampler exists
#==================================================

def check_sampler_exist(sampler_file, silent=False):
    """
    Check if the sampler already exists
        
    Parameters
    ----------
    - sampler_file (str): the filename of the sampler
    - silent (bool): to show or not info

    Output
    ------
    sampler_exist (bool): true if sampler exist, false otherwise

    """

    sampler_exist = os.path.exists(sampler_file)
    
    if not silent:
        if sampler_exist:
            print('----- Existing sampler:')
            print('      '+sampler_file)
        else:
            print('----- No existing sampler found')

    return sampler_exist


#==================================================
# Define backend
#==================================================

def define_emcee_backend(sampler_file, sampler_exist, mcmc_reset, nwalkers, ndim, silent=False):
    """
    Define the backend used for emcee
        
    Parameters
    ----------
    - sampler_file (str): the filename of the sampler
    - silent (bool): to show or not info

    Output
    ------
    backend (emcee object): the backend to be used for mcmc run

    """

    backend = emcee.backends.HDFBackend(sampler_file)
        
    if not silent: print('----- Does the sampler already exist? -----')
    if sampler_exist:
        if mcmc_reset:
            if not silent: print('      - Yes, but reset the MCMC even though the sampler already exists')
            backend.reset(nwalkers, ndim)
        else:
            if not silent: print('      - Yes, use the existing MCMC sampler')
            if not silent: print("      - Initial size: {0}".format(backend.iteration))
    else:
        print('      - No, start from scratch')
        backend.reset(nwalkers, ndim)
        
    return backend


#==================================================
# Extract best-fit parameters
#==================================================

def get_emcee_bestfit_param(sampler,
                            burnin=0):
    """
    Extract the best fit parameters from the sampler
        
    Parameters
    ----------
    - sampler (emcee sampler): the sampler
    - burnin (int): remove the first samples of the chains

    Output
    ------
    par_best (list): the best_fit parameters

    """

    param_chains = sampler.chain[:, burnin:, :]
    lnL_chains   = sampler.lnprobability[:, burnin:]
    par_flat = param_chains.reshape(param_chains.shape[0]*param_chains.shape[1], param_chains.shape[2])
    lnL_flat = lnL_chains.reshape(lnL_chains.shape[0]*lnL_chains.shape[1])
    wbest = np.where(lnL_flat == np.amax(lnL_flat))[0][0]
    par_best = par_flat[wbest]

    return par_best


#==================================================
# Extract random parameters from the chains
#==================================================

def get_emcee_random_param(sampler,
                           burnin=0,
                           Nmc=100):
    """
    Extract Nmc random set of parameter from the chain
        
    Parameters
    ----------
    - sampler (emcee sampler): the sampler
    - burnin (int): remove the first samples of the chains
    - Nmc (int): number of parameter set to extract

    Output
    ------
    param (list): the parameters grid (Nmc x Nparam)

    """

    param_chains = sampler.chain[:, burnin:, :]
    par_flat = param_chains.reshape(param_chains.shape[0]*param_chains.shape[1], param_chains.shape[2])
    loc = np.random.randint(0, par_flat.shape[0]-1, size=Nmc)
    pars = par_flat[loc,:]
    
    return pars


#==================================================
# Compute chain statistics
#==================================================

def chains_statistics(param_chains,
                      lnL_chains,
                      parname=None,
                      conf=68.0,
                      show=True,
                      outfile=None):
    """
    Get the statistics of the chains, such as maximum likelihood,
    parameters errors, etc.
        
    Parameters
    ----------
    - param_chains (np array): parameters as Nchain x Npar x Nsample
    - lnl_chains (np array): log likelihood values corresponding to the chains
    - parname (list): list of parameter names
    - conf (float): confidence interval in %
    - show (bool): show or not the values
    - outfile (str): full path to file to write results

    Output
    ------
    - par_best (float): best-fit parameter
    - par_percentile (list of float): median, lower bound at CL, upper bound at CL
    
    """
    
    if outfile is not None:
        file = open(outfile,'w')
        
    Npar = len(param_chains[0,0,:])

    wbest = (lnL_chains == np.amax(lnL_chains))
    par_best       = np.zeros(Npar)
    par_percentile = np.zeros((3, Npar))
    for ipar in range(Npar):
        # Maximum likelihood
        par_best[ipar]          = param_chains[:,:,ipar][wbest][0]

        # Median and xx % CL
        perc = np.percentile(param_chains[:,:,ipar].flatten(),
                             [(100-conf)/2.0, 50, 100 - (100-conf)/2.0])
        par_percentile[:, ipar] = perc
        if show:
            if parname is not None:
                parnamei = parname[ipar]
            else:
                parnamei = 'no name'

            q = np.diff(perc)
            txt = "{0}_{{-{1}}}^{{{2}}}"
            txt = txt.format(perc[1], q[0], q[1])
            
            medval = str(perc[1])+' -'+str(perc[1]-perc[0])+' +'+str(perc[2]-perc[1])
            bfval = str(par_best[ipar])+' -'+str(par_best[ipar]-perc[0])+' +'+str(perc[2]-par_best[ipar])

            print('param '+str(ipar)+' ('+parnamei+'): ')
            print('   median   = '+medval)
            print('   best-fit = '+bfval)
            print('   '+parnamei+' = '+txt)

            if outfile is not None:
                file.write('param '+str(ipar)+' ('+parnamei+'): '+'\n')
                file.write('  median = '+medval+'\n')
                file.write('  best   = '+bfval+'\n')
                file.write('   '+parnamei+' = '+txt+'\n')

    if outfile is not None:
        file.close() 
            
    return par_best, par_percentile




"""
This file contain a subclass of the main_model.py module and ClusterModel class. It
is dedicated to the computing of mock observables.

"""

#==================================================
# Requested imports
#==================================================

import numpy as np
from scipy.interpolate import interp1d
from astropy.coordinates import SkyOffsetFrame, SkyCoord
import astropy.units as u
import astropy.constants as cst
from scipy.spatial.transform import Rotation

from pitszi import utils
from pitszi import utils_pk


#==================================================
# Mock class
#==================================================

class ModelMock(object):
    """ ModelMock class
    This class serves as a parser to the main Model class, to
    include the subclass ModelMock in this other file.

    Attributes
    ----------
    The attributes are the same as the Model class, see model_main.py

    Methods
    ----------
    - get_pressure_profile
    - get_pressure_fluctuation_spectrum
    - get_pressure_cube_profile
    - get_pressure_cube_fluctuation
    - get_sz_map
    
    """
    
    #==================================================
    # Get the electron pressure profile
    #==================================================

    def get_pressure_profile(self,
                             radius=np.logspace(0,4,100)*u.kpc):
        """
        Get the thermal electron pressure profile.
        
        Parameters
        ----------
        - radius (quantity) : the physical 3d radius in units homogeneous to kpc, as a 1d array

        Outputs
        ----------
        - radius (quantity): the 3d radius in unit of kpc
        - p_r (quantity): the electron pressure profile in unit of keV cm-3

        """
        
        # In case the input is not an array
        radius = utils.check_qarray(radius, unit='kpc')

        # get profile
        p_r = self._get_generic_profile(radius, self._model_pressure_profile)
        p_r[radius > self._R_truncation] *= 0

        return radius, p_r.to('keV cm-3')


    #==================================================
    # Get the electron pressure fluctuation spectrum
    #==================================================

    def get_pressure_fluctuation_spectrum(self,
                                          kvec=np.logspace(-1,2,1000)*u.Mpc**-1,
                                          kmin_norm=None,
                                          kmax_norm=None,
                                          Npt_norm=10000):
        """
        Get the thermal electron pressure fluctuation spectrum (delta P / P).
        
        Parameters
        ----------
        - kvec (quantity) : the physical wavenumber in units homogeneous to kpc-1, as a 1d array
        - kmin/max_norm (quantity): the wavenumber range used for notmalization
        - Npt_norm (int): number of point to generagte the spectrum used for normalization

        Outputs
        ----------
        - kvec (quantity): the physical wavenumber in units homogeneous to kpc-1, as a 1d array
        - P3d_k (quantity): the 3D power spectrum in unit homogeneous to kpc^3

        """

        # In case the input is not an array
        kvec = utils.check_qarray(kvec, unit='kpc-1')
        
        # compute the model
        '''
        This part could go, as for the profile, in a generic fluctuation library and be read fom there
        This can be done when the library includes more functionnal forms
        '''
        if self._model_pressure_fluctuation['name'] == 'CutoffPowerLaw':

            if kmin_norm is None:
                kmin = 1/(4*self._model_pressure_fluctuation['Linj'].to_value('kpc'))
            else:
                kmin = kmin_norm.to_value('kpc')
            if kmax_norm is None:
                kmax = 4/self._model_pressure_fluctuation['Ldis'].to_value('kpc')
            else:
                kmax = kmax_norm.to_value('kpc')

            # K array for normalization
            kvec_norm = np.logspace(np.log10(kmin),np.log10(kmax), Npt_norm) # kpc

            # First extract the normalization givenintegrating within some k range
            cut_low  = np.exp(-(1/(kvec_norm*self._model_pressure_fluctuation['Linj'].to_value('kpc'))**2))
            cut_high = np.exp(-(kvec_norm*self._model_pressure_fluctuation['Ldis'].to_value('kpc'))**2)
            pl = kvec_norm**self._model_pressure_fluctuation['slope']
            f_k = pl * cut_high * cut_low
            Normalization = utils.trapz_loglog(4*np.pi*kvec_norm**2 * f_k, kvec_norm)

            # Then compute Pk at requested scales accounting for the normalization
            k = kvec.to_value('kpc-1')
            A = self._model_pressure_fluctuation['Norm']
            cut_low  = np.exp(-(1/(k*self._model_pressure_fluctuation['Linj'].to_value('kpc'))**2))
            cut_high = np.exp(-(k*self._model_pressure_fluctuation['Ldis'].to_value('kpc'))**2)
            pl = k**self._model_pressure_fluctuation['slope']
            P3d_k = A**2 * pl*cut_high*cut_low / Normalization # adu * u1 / (u1*k**3) = kpc^3
        else:
            raise ValueError("No other model implemented yet")
       
        return kvec, P3d_k*u.kpc**3
    
    
    #==================================================
    # Pressure profile to grid
    #==================================================
    
    def get_pressure_cube_profile(self):
        """
        Fill the grid with the pressure profile.
        
        
        Parameters
        ----------
        
        Outputs
        ----------
        p_grid (ndarray quantity) : the output pressure grid with keV / cm3 unit attached
        
        """

        #----- Get the grid properties
        Nx, Ny, Nz, proj_reso, proj_reso, los_reso = self.get_3dgrid()

        #----- Get the offset of the cluster
        map_center = self.get_map_center()
        offset = (self._coord).transform_to(SkyOffsetFrame(origin=map_center))
        dRA    = -offset.data.lon.to_value('radian')
        dDec   = offset.data.lat.to_value('radian')
        dx     = dRA*self._D_ang.to_value('kpc')   # Offset map along x
        dy     = dDec*self._D_ang.to_value('kpc')  # Offset map along y
        
        #----- Compute the pixel center coordinates
        ctr_xpix_vec = np.linspace(-Nx*proj_reso/2, Nx*proj_reso/2, Nx) - dx
        ctr_ypix_vec = np.linspace(-Ny*proj_reso/2, Ny*proj_reso/2, Ny) - dy
        ctr_zpix_vec = np.linspace(-Nz*los_reso/2,  Nz*los_reso/2,  Nz)

        coord_z, coord_y, coord_x = np.meshgrid(ctr_zpix_vec, ctr_ypix_vec, ctr_xpix_vec,
                                                indexing='ij')

        #----- Apply ellipticity
        angle_1 = self._triaxiality['euler_angle1'].to_value('deg')
        angle_2 = self._triaxiality['euler_angle2'].to_value('deg')
        angle_3 = self._triaxiality['euler_angle3'].to_value('deg')

        rot = Rotation.from_euler('ZXZ', [angle_1, angle_2, angle_3], degrees=True)
        coordinates = np.array([coord_x.flatten(), coord_y.flatten(), coord_z.flatten()]).T
        rotated_coordinates = rot.apply(coordinates)

        rot_coord_x = rotated_coordinates[:,0].reshape(coord_x.shape) / self._triaxiality['min_to_maj_axis_ratio']
        rot_coord_y = rotated_coordinates[:,1].reshape(coord_y.shape) / self._triaxiality['int_to_maj_axis_ratio']
        rot_coord_z = rotated_coordinates[:,2].reshape(coord_z.shape)
    
        #----- Compute the radius grid and flatten
        rad_grid = np.sqrt((rot_coord_x)**2 + (rot_coord_y)**2 + (rot_coord_z)**2)
        rad_grid_flat = rad_grid.flatten()

        '''
        To Do :
        Evaluate log P(r) for r at the nodes of the grid. Then interpolate at the center using 
        scipy.interpolate.interpn
        '''

        #----- Get the pressure profile in 1d
        rad = utils.sampling_array(self._Rmin, self._R_truncation, NptPd=100, unit=True)
        rad, p_rad = self.get_pressure_profile(rad)
        p_rmin = self.get_pressure_profile(self._Rmin)[1]
        
        #----- Interpolate the pressure to the flattened grid
        itpl = interp1d(rad.to_value('kpc'), p_rad.to_value('keV cm-3'),
                        kind='linear', fill_value='extrapolate')
        p_grid_flat = itpl(rad_grid_flat)

        #----- Truncate min and max radii
        p_grid_flat[rad_grid_flat <= self._Rmin.to_value('kpc')] = p_rmin.to_value('keV cm-3')
        p_grid_flat[rad_grid_flat > self._R_truncation.to_value('kpc')] = 0

        #----- Warning
        if self._silent is False:
            if self._R_truncation.to_value('kpc') < np.amax(rad_grid_flat):
                print('----- WARNING: the truncation radius is smaller than the box size.')
                print('               This induces a discountinuity wich may cause ringing.')
        
        #----- Reshape to the grid
        p_grid = p_grid_flat.reshape(Nz, Ny, Nx)
        
        return p_grid*u.keV*u.cm**-3


    #==================================================
    # Pressure fluctuation to grid
    #==================================================
    
    def get_pressure_cube_fluctuation(self,
                                      seed=None,
                                      kmin_input=None,
                                      kmax_input=None,
                                      force_isotropy=False,
                                      Npt=1000):
        """
        Fill the grid with pressure fluctuation, i.e. delta P / P(r)
        
        Parameters
        ----------
        - seed (bool): set to a number for reproducible fluctuations
        - kmin_input/kmax_input (flaot): the min and max k range, in kpc,
        used for sampling the power spectrum. If None, a default value based on 
        thre model is used
        - force_isotropy (bool): set to true to remove non isotropic k modes
        - Npt (int): the number of points used to sample the power spectrum

        Outputs
        ----------
        - fluctuation_cube (np.ndarray) : the output grid of delta P / P(r)
        
        """
        
        #----- Set a seed
        np.random.seed(seed)

        #----- Get the grid properties
        Nx, Ny, Nz, proj_reso, proj_reso, los_reso = self.get_3dgrid()

        #----- Bypass if no fluctuations
        if self._model_pressure_fluctuation['name'] == 'CutoffPowerLaw':
            if self._model_pressure_fluctuation['Norm'] == 0:
                fluctuation_cube = np.zeros((Nz, Ny, Nx))
                return fluctuation_cube

        #----- Get the k arrays along each axis
        k_x = np.fft.fftfreq(Nx, proj_reso) # 1/kpc
        k_y = np.fft.fftfreq(Ny, proj_reso)
        k_z = np.fft.fftfreq(Nz, los_reso)

        #----- Define the k grid and norm vector
        k3d_z, k3d_y, k3d_x = np.meshgrid(k_z, k_y, k_x, indexing='ij')
        k3d_norm = np.sqrt(k3d_x**2 + k3d_y**2 + k3d_z**2)
    
        #----- Get the Pk model on the grid
        k3d_norm_flat = k3d_norm.flatten()                      # Flatten the k3d array
        idx_sort = np.argsort(k3d_norm_flat)                    # Get the sorting index
        revidx = np.argsort(idx_sort)                           # Index to invert rearange by sorting
        k3d_norm_flat_sort = np.sort(k3d_norm_flat)             # Sort the k array
        _, P3d_flat_sort = self.get_pressure_fluctuation_spectrum(k3d_norm_flat_sort[1:]*u.kpc**-1)
        P3d_flat_sort = P3d_flat_sort.to_value('kpc3')          # Take the unit out
        P3d_flat_sort = np.append(np.array([0]), P3d_flat_sort) # Add P(k=0) = 0 back
        P3d_flat = P3d_flat_sort[revidx]                        # Unsort
        P3d_k_grid = P3d_flat.reshape(Nz,Ny,Nx)                 # Unflatten to k cube

        #----- Convert for lognormal
        if self._model_pressure_fluctuation['statistics'] == 'lognormal':
            P3d_k_grid = utils_pk.convert_pkln_to_pkgauss(P3d_k_grid, proj_reso, proj_reso, los_reso)
        
        #----- kill the unwanted mode: zero level and values beyond isotropic range if requested
        kmax_isosphere = self.get_kmax_isotropic().to_value('kpc-1')

        if force_isotropy:
            if not self.silent:
                print('            Non isotropic modes k>',kmax_isosphere,' kpc^-1 are set to zero.')
            P3d_k_grid[k3d_norm > kmax_isosphere] = 0

        #----- Compute the amplitude
        amplitude =  np.sqrt(P3d_k_grid / (proj_reso*proj_reso*los_reso))

        #----- Generate the 3D field
        field = np.random.normal(loc=0, scale=1, size=(Nz,Ny,Nx))
        fftfield = np.fft.fftn(field) * amplitude
        fluctuation_cube = np.real(np.fft.ifftn(fftfield))

        if self._model_pressure_fluctuation['statistics'] == 'lognormal':
            sig2 = np.mean(amplitude**2) # The mean of LN distrib is exp(mu+s^2/2) and sig^2 = np.mean(ampli^2)
            fluctuation_cube -= sig2/2   # So we want the mean to be -(mu + sig^2/2), with mu=0
            fluctuation_cube = np.exp(fluctuation_cube) # take exponential
            fluctuation_cube -= 1                       # Subtract 1 to have dP/P instead of dP/P+1 which is LN
        
        if not self.silent:
            print('----- INFO: fluctuation cube rms.')
            print('            Expected rms over the full k range:', self._model_pressure_fluctuation['Norm'])
            print('            Expected rms given the missing k range:', np.sqrt(np.mean(amplitude**2)))
            print('            Actual rms for this noise realization:', np.std(fluctuation_cube))

        return fluctuation_cube
        

    #==================================================
    # Compute SZ map
    #==================================================
    
    def get_sz_map(self,
                   seed=None,
                   no_fluctuations=False,
                   force_isotropy=False,
                   irfs_convolution_beam=None,
                   irfs_convolution_TF=None):
                   
        """
        Compute the SZ mock image.
        
        Parameters
        ----------
        - seed (bool): set to a number for reproducible fluctuations
        - no_fluctuations (bool): set to true when the pure spherical model is requested
        - force_isotropy (bool): set to true to remove non isotropic k modes
        - irfs_convolution_beam (quantity): PSF FWHM in unit homogeneous to arcsec. If given, 
        the model is convolved with the beam
        - irfs_convolution_TF (dictionary): if given, the model is convolved 
        with the transfer function       

        Outputs
        ----------
        - compton (np.ndarray) : the map in units of Compton parameter
        
        """

        #----- Get P(r) grid
        pressure_profile_cube = self.get_pressure_cube_profile()

        #----- Get delta P(x,y,z) grid
        if no_fluctuations:
            pressure_fluctuation_cube = pressure_profile_cube.value*0
        else:
            pressure_fluctuation_cube = self.get_pressure_cube_fluctuation(seed=seed, force_isotropy=force_isotropy)

        #----- Go to Compton
        intPdl = np.sum(pressure_profile_cube*(1 + pressure_fluctuation_cube), axis=0) * self._los_reso
        compton = (cst.sigma_T / (cst.m_e * cst.c**2) * intPdl).to_value('')
        
        #----- Convolution with instrument response
        if (irfs_convolution_beam is not None) or (irfs_convolution_TF is not None):
            if irfs_convolution_beam is not None:
                apps_beam = True
                psf_fwhm = irfs_convolution_beam.to_value('arcsec')
            else:
                apps_beam = False
                psf_fwhm = 0

            if irfs_convolution_TF is not None:
                apps_TF_LS = True
                TF = irfs_convolution_TF
            else:
                apps_TF_LS = False
                kref = np.linspace(0, 1, 1000)*u.arcsec**-1
                TF = {'k':kref, 'TF':kref.value*0+1}
                
            compton = utils_pk.apply_transfer_function(compton,
                                                       self.get_map_reso().to_value('arcsec'),
                                                       psf_fwhm, TF,
                                                       apps_TF_LS=apps_TF_LS, apps_beam=apps_beam)
            
        return compton


 



"""
This file contain the Physics class. 
It is dedicated to derive physical constraints on the 
cluster given the Pk3d inference.
"""

#==================================================
# Requested imports
#==================================================

import os
import numpy as np
import astropy.units as u
import astropy.constants as cst
import astropy.cosmology
from scipy.ndimage import gaussian_filter
import scipy.stats as stats
from scipy.optimize import brentq
import pprint
import copy
import pickle
import dill

from minot.ClusterTools import cluster_global
from pitszi import title


#==================================================
# Mean and std for lognormal distribution
#==================================================

def stat_from_paramLN(mu, sigma):
    """
    Compute the mean and standard deviation of a variable that follows 
    a lognormal distribution:

    f(X) = 1/(X sigma sqrt(2 pi)) . exp( -(ln X - mu)^2 / (2 sigma^2))
        
    In practice, ln (X == delta P / P + 1) follows a normal distribution, 
    centered on mu=0.

    Parameters
    ----------
    - mu (quantity): gaussian mean parameter
    - sigma (float): gaussian sigma parameter
            
    Outputs
    ----------
    - mean (float): mean of the variable
    - std (float): standard deviation of the variable

    """

    mean = np.exp(mu + sigma**2/2)
    std = ((np.exp(sigma**2) - 1) * np.exp(2*mu + sigma**2))**0.5

    return mean, std


#==================================================
# Sigma and mu parameter from mean and std of lognormal variable
#==================================================

def paramLN_from_stat(mean, std):
    """
    Compute the sigma and mu parameter of a lognormal distribution
    given the mean and standard deviation of the parameter X, for which ln X 
    is normally distributed.
        
    Parameters
    ----------
    - mean (float): mean of the varible that follow LN PDF
    - std (float): std of the variable that follow LN PDF
            
    Outputs
    ----------
    - mu (float): mu parameter of the lognormal function
    - sigma (float): sigma parameter of the lognormal function

    """

    sigma = (np.log(1 + std**2 / mean**2))**0.5
    mu    = np.log(mean) - 0.5*sigma**2

    return mu, sigma


#==================================================
# Std of lognormal distribution from sigma parameter
#==================================================

def std_from_sigmaLN(sigma):
    """
    Compute the standard deviation of a variable that follows 
    a lognormal distribution with mu=0:

    f(X) = 1/(X sigma sqrt(2 pi)) . exp( -(ln X)^2 / (2 sigma^2))
        
    In practice, ln (X == delta P / P + 1) follows a normal distribution, 
    centered on mu=0.

    Parameters
    ----------
    - sigma (float): gaussian sigma parameter
            
    Outputs
    ----------
    - std (float): standard deviation of X, followed by the lognormal variable

    """
    return ((np.exp(sigma**2) - 1) * np.exp(2*mu + sigma**2))**0.5


#==================================================
# Sigma parameter of lognormal distribution from std
#==================================================

def sigmaLN_from_std(std):
    """
    Compute the sigma parameter of a lognormal distribution with mu=0
    given the standard deviation of the parameter X, for which ln X 
    is normally distributed. 
        
    Parameters
    ----------
    - std (float): gaussian sigma parameter
            
    Outputs
    ----------
    - sigma (float): sigma parameter of the lognormal function

    """
    return np.sqrt(np.log((1+np.sqrt(1+4*std**2))/2))


#==================================================
# Sound speed
#==================================================

def sound_speed(kbT, gamma=5.0/3, mu_g=0.61):
    """
    Compute the sound speed given the temperature.
        
    Parameters
    ----------
    - kbT (quantity): the gas temperature homogeneous to keV. 
    Can be a np array.
    - gamma (float): adiabatic index
    - mu_g (float): mean molecular weight of the gas
            
    Outputs
    ----------
    - c_s (quantity): sound speed homogeneous to km/s

    """

    c_s = np.sqrt((gamma*kbT)/(mu_g*cst.m_p))
    
    return c_s.to('km/s')


#==================================================
# Compute sigma_v
#==================================================

def sigma_v(c_s, M3d):
    """
    Compute sigma_v = sound speed x 3D mach number

    Parameters
    ----------
    - c_s (quantity): sound speed homogeneous to km/s
    - M3d(float): 3D Mach number
            
    Outputs
    ----------
    - sigma_v (quantity): homogeneous to km/s

    """

    sigma_v = c_s * M3d
    
    return sigma_v.to('km/s')


#==================================================
# Turbulent to thermal energy ratio
#==================================================

def Eturb_over_Etherm(M3d, gamma=5/3.0):
    """
    Compute turbulent to thermal energy ratio from Mach number.
    This is also the turbulent to thermal pressure ratio

    Eturb = 1/2 rho_gas sigma_v**2
          = 1/2 rho_gas c_s**2 * M3d**2
          = 1/2 rho_gas gamma P_gas / rho_gas * M3d**2
          = 1/2 gamma (gamma-1) Eth * M3d**2
    ==> Eturb / Eth = 1/2 gamma (gamma-1) M3d**2
    
    Parameters
    ----------
    - M3d(float): 3D Mach number
    - gamma (float): adiabatic index
            
    Outputs
    ----------
    - X (float): turbulent to thermal energy ratio

    """

    X = 0.5 * gamma*(gamma-1) * M3d**2
    
    return X


#==================================================
# Mach number from A(Linj) adiabatic Gaspari & Churazov (2013)
#==================================================

def Apeak_to_M3d_G13(A_kpeak, Linj=500*u.kpc, alpha=0.25, gamma=5/3.0):
    """
    Gaspari and Churazov 2013 A&A 559, A78 (2013)
    Implement the 3D Mach number from the pressure 
    fluctuation power spectrum peak amplitude 
    This uses the case of density perturbations and 
    assumes adiabatic perturbations 
    delta P/P = gamma delta n / n
        
    A_rho = 0.25 M3d (Linj/500)^alpha_H (l/Linj)^alpha_c

    Parameters
    ----------
    - A_kpeak (float): the characteristic amplitude of the power spectrum 
    taken at the peak: sqrt(4 pi k_peak^3 P(k_peak)) ; k_peak = 1/Linj
    - Linj (quantity): injection scale, homogeneous to kpc
    - alpha (float): dependence with Linj (~0.2-0.3), see Gaspari & Churazov (2013)
    - gamma (float): adiabatic index
            
    Outputs
    ----------
    - M3d(float): 3D Mach number

    """

    scale = 1/(0.25*gamma)
    M3d = scale * A_kpeak * (500 / Linj.to_value('kpc'))**alpha
    
    return M3d


#==================================================
# Mach number from \int Pk from Mohapatra+
#==================================================

def sigma_to_M3d_M20(sigma, gamma=5.0/3):
    """
    Rajsekhar Mohapatra +2020 MNRAS 493, 5838–5853 (2020) 
    Rajsekhar Mohapatra +2021 MNRAS 500, 5072–5087 (2021)
    Rajsekhar Mohapatra +2022 MNRAS 510, 3778–3793 (2022)
    Implement the result from their scaling (Figure 6) for the pressure.

    Parameters
    ----------
    - sigma (float): the square root of the integral of Pk
    sigma^2 = \int 4 pi k^2 Pk dk. sigma is directly the norm of 
    pitszi power spectrum
    - gamma (float): adiabatic index

    Outputs
    ----------
    - M3d(float): 3D Mach number

    """

    # b = 1/3 for solenoidal forcing
    # 
    
    # Convert to what would be the sigma of a lognormal pdf for delta P/P + 1
    sigmaLN = sigmaLN_from_std(sigma)

    # Apply the scaling (eq 22)
    M3d = ((np.exp(sigmaLN**2) - 1) / ((1/3.0)**2 * gamma**2))**(1./4)
    
    return M3d


#==================================================
# Mach number from \int Pk from Zhuravleva + 2023
#==================================================

def sigma_to_M3d_Z23(sigma, state='in-between', ell=True, ret_err=False):
    """
    Zhuravleva + 2023 (MNRAS 520, 5157–5172 (2023))
    Implement the result from their Figure 6 for the pressure.
    See also Zhuravleva+ MNRAS 428, 3274–3287 (2013) 

    Parameters
    ----------
    - sigma (float): the sqaure root of the integral of Pk
    sigma^2 = \int 4 pi k^2 Pk dk. sigma is directly the norm of 
    pitszi power spectrum
    - state (str): 'relaxed', 'in-between', 'unrelaxed' according to 
    the table in Zhuravleva + 2023
    - ell (bool): if true, take the elliptical case from Zhuravleva + 2023,
    otherwise take the spherical case
    - ret_err (bool): set to true to return errors
    
    Outputs
    ----------
    - M3d(float): 3D Mach number
    - M3d_err(float): 3D Mach number error, from simple error propagation

    """

    # Convert to what would be the sigma of a lognormal pdf for delta P/P + 1
    sigmaLN = sigmaLN_from_std(sigma)

    # Apply the definition of Zhuravleva+23
    dxi_xi = 2*np.sqrt(2*np.log(2)) / np.log(10) * sigmaLN

    # Apply the scaling for different cases
    if ell: #----- Elliptical case
        if state == 'relaxed':
            M1d     = dxi_xi / 1.0 
            M1d_err = dxi_xi / 1.0**2 * 0.2
        elif state == 'in-between':
            M1d     = dxi_xi / 1.2
            M1d_err = dxi_xi / 1.2**2 * 0.4
        elif state == 'unrelaxed':        
            M1d     = dxi_xi / 1.5
            M1d_err = dxi_xi / 1.5**2 * 0.5
        else:
            raise ValueError('Only relaxed, in-between and unrelaxed states are available')
        
    else: #----- Spherical case
        if state == 'relaxed':
            M1d     = dxi_xi / 1.4
            M1d_err = dxi_xi / 1.4**2 * 0.3
        elif state == 'in-between':
            M1d     = dxi_xi / 1.4
            M1d_err = dxi_xi / 1.4**2 * 0.3
        elif state == 'unrelaxed':        
            M1d     = dxi_xi / 1.5
            M1d_err = dxi_xi / 1.5**2 * 0.3
        else:
            raise ValueError('Only relaxed, in-between and unrelaxed states are available')

    # Convert to 3d Mach number
    M3d     = np.sqrt(3)*M1d
    M3d_err = np.sqrt(3)*M1d_err

    # Return
    if ret_err:
        return M3d, M3d_err
    else:
        return M3d


#==================================================
# Mach number from \int Pk from Zhuravleva + 2023 Table1
#==================================================

def sigma_rad_to_M3d_Z23(sigma, r, R500, state='in-between', case='prod', ret_err=False):
    """
    Zhuravleva + 2023 (MNRAS 520, 5157–5172 (2023))
    Implement the result from Tab. 1, in the case both radial dependence
    and fluctuations are considered.
    See also Zhuravleva+ MNRAS 428, 3274–3287 (2013) 

    Parameters
    ----------
    - sigma (float): the sqaure root of the integral of Pk
    sigma^2 = \int 4 pi k^2 Pk dk. sigma is directly the norm of 
    pitszi power spectrum
    - r (quantity): radius homogeneous to kpc
    - R500 (quantity): R500 homogeneous to kpc
    - state (str): 'relaxed', 'in-between', 'unrelaxed' according to 
    the table in Zhuravleva + 2023
    - case (str): 'sum' (M1d = alpha + beta r/r500 + gamma dxi/xi) 
    or 'prod' (M1d = alpha (r/r500)^beta (dxi/xi)^gamma)
    - ret_err (bool): set to true to return errors

    Outputs
    ----------
    - M3d (array): 3D Mach number as a function of radius

    """

    r_scaled = r.to_value('kpc') / R500.to_value('kpc')

    # Convert to what would be the sigma of a lognormal pdf for delta P/P + 1
    sigmaLN = sigmaLN_from_std(sigma)

    # Apply the definition of Zhuravleva+23
    dxi_xi = 2*np.sqrt(2*np.log(2)) / np.log(10) * sigmaLN

    #----- Define the parameters
    # Case sum, see Tab. 1
    if case == 'sum':
        if state == 'relaxed':
            a, b, c = 0.06, 0.11, 0.35
            rms = 0.04
        elif state == 'in-between':
            a, b, c = 0.13, 0.11, 0.24
            rms = 0.07
        elif state == 'unrelaxed':
            a, b, c =  0.21, 0.14, 0.14
            rms = 0.07
        else:
            raise ValueError('Only relaxed, in-between and unrelaxed states are available')

        M3d     = (a + b*r_scaled + c*dxi_xi) *np.sqrt(3)
        M3d_err = M3d * 0 + rms
        
    # Case prod, see Tab. 1
    elif case == 'prod':
        if state == 'relaxed':
            a, b, c = 0.4 , 0.42, 0.29
            rms = 0.04
        elif state == 'in-between':
            a, b, c = 0.45, 0.26, 0.28
            rms = 0.07
        elif state == 'unrelaxed':
            a, b, c = 0.50, 0.21, 0.27
            rms = 0.07
        else:
            raise ValueError('Only relaxed, in-between and unrelaxed states are available')

        M3d     = a * r_scaled**b * dxi_xi**c * np.sqrt(3)
        M3d_err = M3d * 0 + rms

    else:
        raise ValueError('Only sum and prod are possible cases')

    if ret_err:
        return M3d, M3d_err
    else:
        return M3d

    
#==================================================
# Expected NT pressure profile: Shaw et al (2010)
#==================================================

def pnt_over_ptot_r_S10(r, R500, z):
    """
    Implement the nonthermal to thermal pressure from 
    L. Shaw + 2010 (The Astrophysical Journal, 725:1452–1465, 2010 December 20)
    See their Eq 16

    Parameters
    ----------
    - r (quantity): radius homogeneous to kpc
    - R500 (quantity): R500 homogeneous to kpc
    - z (float): cluster redshift

    Outputs
    ----------
    - Pnt_Ptot (array): the ratio between nonthermal to total pressure as a function of radius

    """

    r_scale500 = r.to_value('kpc') / R500.to_value('kpc')
    
    alpha_0 = 0.18 # pm 0.06
    n_nt    = 0.8  # pm 0.25
    beta    = 0.5
    
    fmax = 4**(-n_nt)/alpha_0
    comp1 = (1+z)**beta
    comp2 = (fmax-1)*np.tanh(beta*z)+1
    f_z = np.amin(np.array([comp1, comp2]))
    alpha_z = alpha_0 * f_z
    
    Pnt_Ptot = alpha_z * r_scale500**n_nt
    
    return Pnt_Ptot


#==================================================
# Expected NT pressure profile: Battaglia et al (2011)
#==================================================

def pnt_over_ptot_r_B11(r, R500, z, c500=3.0, cosmo=astropy.cosmology.Planck15):
    """
    Implement the nonthermal to thermal pressure from 
    N. Battaglia + 2012 (The Astrophysical Journal, 758:74 (23pp), 2012 October 20)
    See their Eq 16
    
    Parameters
    ----------
    - r (quantity): radius homogeneous to kpc
    - R500 (quantity): R500 homogeneous to kpc
    - z (float): cluster redshift
    - c500 (float): NFW concentration parameter
    - cosmo (astropy.cosmology): cosmology object to use

    Outputs
    ----------
    - Pnt_Ptot (array): the ratio between nonthermal to total pressure as a function of radius

    """

    # convert R500 to get M200
    M500 = cluster_global.Rdelta_to_Mdelta(R500.to_value('kpc'), z, delta=500, cosmo=cosmo)
    M200 = cluster_global.Mdelta1_to_Mdelta2_NFW(M500, delta1=500, delta2=200, c1=c500, redshift=z, cosmo=cosmo)

    # Apply functional form
    r_scale500 = r.to_value('kpc') / R500.to_value('kpc')

    alpha_0 = 0.18
    n_nt    = 0.8
    beta    = 0.5
    alpha_z = alpha_0 * (1+z)**beta
    Pnt_Ptot = alpha_z * r_scale500**n_nt * (M200 / 3e14)**(1.0/5)
    
    return Pnt_Ptot


#==================================================
# Expected NT pressure profile: Nelson et al (2014)
#==================================================

def pnt_over_ptot_r_N14(r, R500, z, c500=3.0, cosmo=astropy.cosmology.Planck15):
    """
    Implement the nonthermal to thermal pressure from 
    Nelson +14 (The Astrophysical Journal, 792:25 (8pp), 2014 September)
    See their Eq 7.
    We need to convert from R500 to R200m assuming NFW profile in the process.

    Parameters
    ----------
    - r (quantity): radius homogeneous to kpc
    - R500 (quantity): R500 homogeneous to kpc
    - z (float): cluster redshift
    - c500 (float): NFW concentration parameter
    - cosmo (astropy.cosmology): cosmology object to use

    Outputs
    ----------
    - Pnt_Ptot (array): the ratio between nonthermal to total pressure as a function of radius

    """
    
    #----- Compute R200m
    # Get M500 critical
    M500c = cluster_global.Rdelta_to_Mdelta(R500.to_value('kpc'), z, delta=500, cosmo=cosmo) # Msun

    # NFW profile
    Rs   = R500.to_value('kpc') / c500      # kpc
    rho0 = M500c / (4*np.pi * Rs**3) / (np.log(1 + c500) - c500/(1+c500)) # Msun/kpc^3
    
    # Compute R200m
    delta = 200
    delta_rho_c = delta * cosmo.critical_density(z).to_value('Msun kpc-3')
    delta_rho_m = delta_rho_c * (cosmo.Om0*(1+z)**3) / (cosmo.Om0*(1+z)**3 + cosmo.Ode0)
    def enclosed_mass_difference(radius):
        mass_nfw    = (4*np.pi * Rs**3) * rho0 * (np.log(1 + radius/Rs) - radius/(radius+Rs))
        vol = 4.0/3*np.pi*radius**3
        return  mass_nfw/vol - delta_rho_m
    
    R200m = brentq(enclosed_mass_difference, R500.to_value('kpc')*1e-5, R500.to_value('kpc')*1e5)
    r_scale = r.to_value('kpc') / R200m
    
    #----- Apply the functional form
    A = 0.452
    B = 0.841
    gamma = 1.628
    Pnt_Ptot = 1 - A * (1+ np.exp(-(r_scale/B)**gamma))
    
    return Pnt_Ptot


#==================================================
# Inference class
#==================================================

class Physics():
    """ Physics class
    This class derives physical constraints from 
    Pk3d sampling.

    Attributes
    ----------
        - something: contain the sampling information

        # Admin
        - silent (bool): set to False to give information
        - output_dir (str): the output directory
        
    Methods
    ----------
    - print_param: print the current inference parameters
    """    

    #==================================================
    # Initialization
    #==================================================

    def __init__(self,
                 something,
                 silent=False,
                 output_dir='./pitszi_output',
    ):
        """
        Initialize the inference object. 
        All parameters can be changed on the fly.

        Parameters
        ----------
        - something: the sampling information
 
        - silent (bool): set to False for printing information
        - output_dir (str): directory where outputs are saved
        
        """

        if not silent:
            title.show_physics()
        
        #----- Somthing that contains sampling information
        self.something = something

        #----- Admin
        self.silent     = silent
        self.output_dir = output_dir

        
    #==================================================
    # Print parameters
    #==================================================
    
    def print_param(self):
        """
        Print the current parameters describing the inference.
        
        Parameters
        ----------
        
        Outputs
        ----------
        The parameters are printed in the terminal
        
        """

        print('=====================================================')
        print('============= Current Inference() state =============')
        print('=====================================================')
        
        pp = pprint.PrettyPrinter(indent=4)
        
        par = self.__dict__
        keys = list(par.keys())
        
        for k in range(len(keys)):
            print(('--- '+(keys[k])))
            print(('    '+str(par[keys[k]])))
            print(('    '+str(type(par[keys[k]]))+''))
        print('=====================================================')


    #==================================================
    # Save parameters
    #==================================================
    
    def save_physics(self):
        """
        Save the current physics object.
        
        Parameters
        ----------
            
        Outputs
        ----------
        The parameters are saved in the output directory

        """

        # Create the output directory if needed
        if not os.path.exists(self.output_dir): os.mkdir(self.output_dir)

        # Save
        with open(self.output_dir+'/physics_parameters.pkl', 'wb') as pfile:
            #pickle.dump(self.__dict__, pfile, pickle.HIGHEST_PROTOCOL)
            dill.dump(self.__dict__, pfile)

        # Text file for user
        par = self.__dict__
        keys = list(par.keys())
        with open(self.output_dir+'/physics_parameters.txt', 'w') as txtfile:
            for k in range(len(keys)):
                txtfile.write('--- '+(keys[k])+'\n')
                txtfile.write('    '+str(par[keys[k]])+'\n')
                txtfile.write('    '+str(type(par[keys[k]]))+'\n')

                
    #==================================================
    # Load parameters
    #==================================================
    
    def load_physics(self, param_file):
        """
        Read the a given parameter file to re-initialize the physics object.
        
        Parameters
        ----------
        param_file (str): the parameter file to be read
            
        Outputs
        ----------
            
        """

        with open(param_file, 'rb') as pfile:
            par = dill.load(pfile)
            
        self.__dict__ = par
        



__version__ = '0.2.0'

try:
    from .model_main import Model
    from .data_main import Data
    from .inference_radial_main import InferenceRadial
    from .inference_fluctuation_main import InferenceFluctuation

except ImportError:
    print('WARNING: Could not import submodules                             ')
    print('     from .model_main import Model                               ')
    print('     from .data_main import Data                                 ')
    print('     from .inference_radial_main import InferenceRadial          ')
    print('     from .inference_fluctuation_main import InferenceFluctuation')
    print('                                                                 ')
    print('         You may try (re)installing dependencies                 ')
    print('         by hand. For example running:                           ')
    print('             $ conda install numpy                               ')
    print('             $ conda install astropy                             ')
    print('             $ conda install scipy                               ')
    print('             $ conda install matplotlib                          ')
    print('             $ conda install emcee                               ')
    print('             $ conda install pandas                              ')
    print('             $ conda install seaborn                             ')
    print('             $ conda install dill                                ')
    print('             $ conda install pathos                              ')
    print('             $ pip install minot                                 ')



"""
This file contain the InferenceRadial class. 
It is dedicated to extract constraints on radial model parameters.
"""

#==================================================
# Requested imports
#==================================================

import os
import pprint
import copy
import pickle
import dill

from pitszi import title
from pitszi.inference_radial_fitting import InferenceRadialFitting


#==================================================
# Inference class
#==================================================

class InferenceRadial(InferenceRadialFitting):
    """ Inference class
        This class infer the radial profile given input
        data (from class Data()) and model (from class Model())
        that are attributes

    Attributes
    ----------
        - data (object from Class Data()): the data object
        - model (object from class Model()): the model object

        # Nuisance parameters
        self.nuisance_ZL (float): nuisance parameter -- the map zero level

        # Analysis methodology
        - method_use_covmat (bool): use covariance matrix in fit?
        - method_parallel (bool): use parallelization in fit

        # MCMC parameters
        - mcmc_nwalkers (int): number of MCMC walkers
        - mcmc_nsteps (int): number of MCMC steps to run
        - mcmc_burnin (int): number of burnin steps
        - mcmc_reset (bool): restart MCMC from scratch even if sampler exists?
        - mcmc_run (bool): run MCMC?
        - mcmc_Nresamp (int): number of resampling model in post-MCMC analysis

        # Admin
        - silent (bool): set to False to give information
        - output_dir (str): the output directory
        
    Methods
    ----------
    - print_param: print the current inference parameters
    - save_inference: save the class object state
    - load_inference: load the class object
    - get_radial_data: get the data map in the case of the radial component (i.e. the y map)
    - get_radial_noise_statistics: get the noise properties in the case of the radail component
    - get_radial_model: compute the model in the case of the radial component
    
    """    

    #==================================================
    # Initialization
    #==================================================

    def __init__(self,
                 data,
                 model,
                 #
                 nuisance_ZL=0,
                 #
                 method_use_covmat=False,
                 method_parallel=False,
                 #
                 mcmc_nwalkers=20,
                 mcmc_nsteps=500,
                 mcmc_burnin=100,
                 mcmc_reset=False,
                 mcmc_run=True,
                 mcmc_Nresamp=100,
                 #
                 silent=False,
                 output_dir='./pitszi_output',
    ):
        """
        Initialize the inference object. 
        All parameters can be changed on the fly.

        Parameters
        ----------
        - data (pitszi Class Data object): the data
        - model (pitszi Class Model object): the model
        - nuisance_ZL (float): map zero level, a nuisance parameter.
        - method_use_covmat (bool): use covariance matrix in the fit
        - method_use_paral (bool): use parallelization      
        - mcmc_nwalkers (int): number of MCMC walkers
        - mcmc_nsteps (int): number of MCMC steps to run
        - mcmc_burnin (int): the burnin, number of point to remove from the chains
        - mcmc_reset (bool): True for restarting the MCMC even if a sampler exists
        - mcmc_run (bool): True to run the MCMC
        - mcmc_Nresamp (int): the number of Monte Carlo to resample the chains for results
        - silent (bool): set to False for printing information
        - output_dir (str): directory where outputs are saved
        
        """

        if not silent:
            title.show_inference_radial()
        
        #----- Input data and model (deepcopy to avoid modifying the input when fitting)
        self.data  = copy.deepcopy(data)
        self.model = copy.deepcopy(model)

        #----- Nuisance parameters
        self.nuisance_ZL     = 0

        #----- Analysis methodology
        self.method_use_covmat = method_use_covmat
        self.method_parallel = method_parallel

        #----- MCMC parameters
        self.mcmc_nwalkers = mcmc_nwalkers
        self.mcmc_nsteps   = mcmc_nsteps
        self.mcmc_burnin   = mcmc_burnin
        self.mcmc_reset    = mcmc_reset
        self.mcmc_run      = mcmc_run
        self.mcmc_Nresamp  = mcmc_Nresamp

        #----- Admin
        self.silent     = silent
        self.output_dir = output_dir

        
    #==================================================
    # Print parameters
    #==================================================
    
    def print_param(self):
        """
        Print the current parameters describing the inference.
        
        Parameters
        ----------
        
        Outputs
        ----------
        The parameters are printed in the terminal
        
        """

        print('=====================================================')
        print('============= Current Inference() state =============')
        print('=====================================================')
        
        pp = pprint.PrettyPrinter(indent=4)
        
        par = self.__dict__
        keys = list(par.keys())
        
        for k in range(len(keys)):
            print(('--- '+(keys[k])))
            print(('    '+str(par[keys[k]])))
            print(('    '+str(type(par[keys[k]]))+''))
        print('=====================================================')


    #==================================================
    # Save parameters
    #==================================================
    
    def save_inference(self):
        """
        Save the current inference object.
        
        Parameters
        ----------
            
        Outputs
        ----------
        The parameters are saved in the output directory

        """

        # Create the output directory if needed
        if not os.path.exists(self.output_dir): os.mkdir(self.output_dir)

        # Save
        with open(self.output_dir+'/inference_parameters.pkl', 'wb') as pfile:
            #pickle.dump(self.__dict__, pfile, pickle.HIGHEST_PROTOCOL)
            dill.dump(self.__dict__, pfile)

        # Text file for user
        par = self.__dict__
        keys = list(par.keys())
        with open(self.output_dir+'/inference_parameters.txt', 'w') as txtfile:
            for k in range(len(keys)):
                txtfile.write('--- '+(keys[k])+'\n')
                txtfile.write('    '+str(par[keys[k]])+'\n')
                txtfile.write('    '+str(type(par[keys[k]]))+'\n')

                
    #==================================================
    # Load parameters
    #==================================================
    
    def load_inference(self, param_file):
        """
        Read the a given parameter file to re-initialize the inference object.
        
        Parameters
        ----------
        param_file (str): the parameter file to be read
            
        Outputs
        ----------
            
        """

        with open(param_file, 'rb') as pfile:
            par = dill.load(pfile)
            
        self.__dict__ = par
        
            
    #==================================================
    # Data for radial fit
    #==================================================
    
    def get_radial_data(self):
        """
        This function returns the data to be compared with the model in the case of radial
        fitting
        
        Parameters
        ----------

        Outputs
        ----------
        - image (2d np array): the data image

        """

        return self.data.image
    
    
    #==================================================
    # Noise for radial fit
    #==================================================
    
    def get_radial_noise_statistics(self):
        """
        This function returns the noise properties in the case of radial
        fitting
        
        Parameters
        ----------

        Outputs
        ----------
        - noise_rms (2d np array): the rms associated with the image
        - noise_covmat (Npix**2 np array): the noise covariance matrix

        """

        noise_rms = self.data.noise_rms
        if noise_rms is None:
            raise ValueError("Noise rms not available in data. Please use the class Data() to define it")

        if self.method_use_covmat:
            noise_covmat = self.data.noise_covmat
            if noise_covmat is None:
                raise ValueError("Noise covariance not available in data. Use the class Data() to define it.")
        else:
            noise_covmat = None    

        return noise_rms, noise_covmat
    
    
    #==================================================
    # Model for radial fit
    #==================================================
    
    def get_radial_model(self):
        """
        This function returns the model to be compared to the data in the case of radial
        fitting
        
        Parameters
        ----------

        Outputs
        ----------
        - model_image (2d np array): the model image

        """

        #----- Get the cluster model
        model_img = self.model.get_sz_map(seed=None, no_fluctuations=True, force_isotropy=False,
                                          irfs_convolution_beam=self.data.psf_fwhm,
                                          irfs_convolution_TF=self.data.transfer_function)
                
        #----- Compute the final model
        model_img = model_img + self.nuisance_ZL
        
        return model_img




"""
This file contains utilities related to plots used in pitszi

"""

import numpy as np
from scipy.interpolate import interp1d
from scipy.ndimage import gaussian_filter
from astropy.wcs import WCS
import astropy.units as u
import matplotlib.pyplot as plt
import corner
import seaborn as sns
import pandas as pd
import warnings
import copy

from pitszi import utils


#==================================================
# Plot MCMC chains
#==================================================

def chains_1Dplots(param_chains,
                   parname,
                   output_file):
    """
    Plot related to MCMC chains
        
    Parameters
    ----------
    - param_chains (np array): parameters as Nchain x Npar x Nsample
    - parname (list): list of parameter names
    - output_file (str): output file name

    Output
    ------
    Plots are saved in the output directory

    """

    Npar = len(param_chains[0,0,:])
    Nchain = len(param_chains[:,0,0])

    # Chains
    fig, axes = plt.subplots(Npar, figsize=(8, 2*Npar), sharex=True)
    for i in range(Npar):
        if Npar>1:
            ax = axes[i]
        else:
            ax = axes
            
        for j in range(Nchain):
            ax.plot(param_chains[j, :, i], alpha=0.5)
        ax.set_xlim(0, len(param_chains[0,:,0]))
        ax.set_ylabel(parname[i])
        ax.grid()
    if Npar>1:
        axes[-1].set_xlabel("step number")
    else:
        axes.set_xlabel("step number")
        
    fig.savefig(output_file)
    plt.close()


#==================================================
# Plots 1D chain histograms
#==================================================

def chains_1Dhist(param_chains,
                  parname,
                  out_files,
                  conf=68.0,
                  par_best=None,
                  truth=None):
    """
    Plot related to MCMC chains
        
    Parameters
    ----------
    - param_chains (np array): parameters as Nchain x Npar x Nsample
    - parname (list): list of parameter names
    - out_files (str): file name list for each param
    - par_best (float): best-fit parameter
    - truth (float): the true known value from input simulation

    Output
    ------
    Plots are saved in the output file

    """
    
    Nbin_hist = 40
    Npar = len(param_chains[0,0,:])
    for ipar in range(Npar):
        if par_best is not None:
            par_best_i = par_best[ipar]
        else:
            par_best_i = None
        if truth is not None:
            truth_i = truth[ipar]
        else:
            truth_i = None
    
        flatchain = param_chains[:,:,ipar].flatten()
        
        seaborn_1d(flatchain.astype("float64"),
                   output_fig=out_files[ipar],
                   ci=conf/100, truth=truth_i, best=par_best_i,
                   label=parname[ipar],
                   gridsize=100, alpha=(0.2, 0.4), 
                   figsize=(10,10), fontsize=12,
                   cols=[('blue','grey', 'orange')])
        plt.close("all")


#==================================================
# Plot corner plots
#==================================================

def chains_2Dplots_corner(param_chains,
                          parname,
                          out_file,
                          smooth=1,
                          Nbin_hist=40,
                          truth=None):
    """
    Plot related to MCMC chains
        
    Parameters
    ----------
    - param_chains (np array): parameters as Nchain x Npar x Nsample
    - parname (list): list of parameter names
    - out_file (str): file where to save plot
    - smooth (float): smoothing lenght
    - Nbin_hist (int): number of bins in histogram
    - truth (list): the true values of the recovered parameters

    Output
    ------
    Plots are saved in the output file

    """

    Npar = len(param_chains[0,0,:])
    Nchain = len(param_chains[:,0,0])
    par_flat = param_chains.reshape(param_chains.shape[0]*param_chains.shape[1], param_chains.shape[2])

    # Corner plot using corner
    figure = corner.corner(par_flat,
                           bins=Nbin_hist,
                           color='k',
                           smooth=smooth,
                           labels=parname,
                           quantiles=(0.16, 0.84),
                           levels=[0.68, 0.95])
    if truth is not None:
        axes = np.array(figure.axes).reshape((Npar, Npar))
        for yi in range(Npar):
            for xi in range(yi):
                ax = axes[yi, xi]
                ax.axvline(truth[xi], color="r")
                ax.axhline(truth[yi], color="r")
                ax.plot(truth[xi], truth[yi], "sr")
    figure.savefig(out_file)
    plt.close("all")


#==================================================
# Plot corner plots with seaborn
#==================================================

def chains_2Dplots_sns(param_chains,
                       parname,
                       out_file,
                       smooth=1,
                       gridsize=100,
                       truth=None):
    """
    Plot related to MCMC chains
        
    Parameters
    ----------
    - param_chains (np array): parameters as Nchain x Npar x Nsample
    - parname (list): list of parameter names
    - out_file (str): file where to save plot
    - smooth (float): smoothing lenght
    - gridsize (int): number of bins in the grid
    - truth (list): the true values of the recovered parameters

    Output
    ------
    Plots are saved in the output file

    """

    Npar = len(param_chains[0,0,:])
    Nchain = len(param_chains[:,0,0])
    par_flat = param_chains.reshape(param_chains.shape[0]*param_chains.shape[1], param_chains.shape[2])

    # Correct scale to have numbers ~1
    power = np.round(np.log10(np.abs(np.mean(par_flat,axis=0))))
    par_flat_scale = par_flat * 10**-power[np.newaxis,:]
    parname_scale = []
    for i in range(len(parname)): parname_scale.append('$10^{'+str(int(-power[i]))+'} '+parname[i]+'$')
    
    # Corner plot using seaborn
    df = pd.DataFrame(par_flat_scale, columns=parname_scale)
    seaborn_corner(df, output_fig=out_file,
                   n_levels=15, cols=[('royalblue', 'k', 'grey', 'Blues')], 
                   ci2d=[0.68, 0.95],
                   smoothing1d=smooth, smoothing2d=smooth, gridsize=gridsize,
                   truth=truth, truth_style='star',
                   limits=None,
                   linewidth=2.0, alpha=(0.1, 0.3, 1.0), figsize=((Npar+1)*3,(Npar+1)*3))
    plt.close("all")
    

#==================================================
# 1D distribution plot function using seaborn
#==================================================

def seaborn_1d(chains, output_fig=None, ci=0.68, truth=None,
               best=None, label=None,
               gridsize=100, alpha=(0.2, 0.4), 
               figsize=(10,10), fontsize=12,
               cols=[('blue','grey', 'orange')]):
    '''
    This function plots 1D distributions of MC chains
    
    Parameters:
    -----------
    - chain (array): the chains sampling the considered parameter
    - output_fig (str): full path to output plot
    - ci (float): confidence interval considered
    - truth (float): the expected truth for overplot
    - best (float list): list of float that contain best fit models to overplot
    - label (str): the label of the parameter
    - gridsize (int): the size of the kde grid
    - alpha (tupple): alpha values for the histogram and the 
    overplotted confidence interval
    - figsize (tupple): the size of the figure
    - fontsize (int): the font size
    - cols (tupple): the colors of the histogram, confidence interval 
    values, and confidence interval filled region

    Output:
    -------
    - Plots
    '''

    # Check type
    if type(chains) is not list:
        chains = [chains]

    # Plot dat
    Ndat = len(chains)            # Number of datasets

    # Make sure there are enough colors
    icol = 0
    while len(cols) < Ndat:
        cols.append(cols[icol])
        icol = icol+1

    fig = plt.figure(0, figsize=(8, 6))
    #----- initial plots of histograms + kde
    for idx, ch in enumerate(chains, start=0):
        sns.histplot(ch, kde=True, kde_kws={'cut':3}, color=cols[idx][0], edgecolor=cols[idx][1], 
                     alpha=alpha[0], stat='density')
    ax = plt.gca()
    ymax = ax.get_ylim()[1]
    
    #----- show limits
    for idx, ch in enumerate(chains, start=0):
        if ci is not None:
            perc = np.percentile(ch, [100 - (100-ci*100)/2.0, 50.0, (100-ci*100)/2.0])
            # Get the KDE line for filling below
            if len(ax.lines) > 0:
                try:
                    xkde = ax.lines[idx].get_xdata()
                    ykde = ax.lines[idx].get_ydata()
                    wkeep = (xkde < perc[0]) * (xkde > perc[2])
                    xkde_itpl = np.append(np.append(perc[2], xkde[wkeep]), perc[0])
                    itpl = interp1d(xkde, ykde)
                    ykde_itpl = itpl(xkde_itpl)
                    perc_max = itpl(perc)
                except:
                    perc_max = perc*0+ymax
                    xkde_itpl = perc*1+0
                    ykde_itpl = ymax*1+0
            else:
                perc_max = perc*0+ymax
                xkde_itpl = perc*1+0
                ykde_itpl = ymax*1+0
            
            ax.vlines(perc[0], 0.0, perc_max[0], linestyle='--', color=cols[idx][1])
            ax.vlines(perc[2], 0.0, perc_max[2], linestyle='--', color=cols[idx][1])
            
            if idx == 0:
                ax.vlines(perc[1], 0.0, perc_max[1], linestyle='-.', label='Median', color=cols[idx][1])
                ax.fill_between(xkde_itpl, 0*ykde_itpl, y2=ykde_itpl, alpha=alpha[1], 
                                color=cols[idx][2], label=str(ci*100)+'% CL')
            else:
                ax.vlines(perc[1], 0.0, perc_max[1], linestyle='-.', color=cols[idx][1])
                ax.fill_between(xkde_itpl, 0*ykde_itpl, y2=ykde_itpl, alpha=alpha[1], color=cols[idx][2])
    
        # Show best fit value                   
        if best is not None:
            if type(best) is not list:
                best = [best]
            ax.vlines(best[idx], 0, ymax, linestyle='-', label='Best-fit', linewidth=2, color=cols[idx][1])
        
    # Show expected value                        
    if truth is not None:
        ax.vlines(truth, 0, ymax, linestyle=':', label='Truth', color='k')
    
    # label and ticks
    if label is not None:
        ax.set_xlabel(label)
        ax.set_ylabel('Probability density')
    ax.set_yticks([])
    ax.set_ylim(0, ymax)
               
    for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] + ax.get_xticklabels() + ax.get_yticklabels()):
        item.set_fontsize(fontsize)
               
    ax.legend(fontsize=fontsize)
    
    if output_fig is not None:
        plt.savefig(output_fig)


#==================================================
# Corner plot function using seaborn
#==================================================

def seaborn_corner(dfs, output_fig=None, ci2d=[0.95, 0.68], ci1d=0.68,
                   truth=None, truth_style='star', labels=None,
                   smoothing1d=1,smoothing2d=1,
                   gridsize=100, linewidth=0.75, alpha=(0.3, 0.3, 1.0), n_levels=None,
                   zoom=1.0/10, add_grid=True,
                   limits=None,
                   figsize=(10,10), fontsize=12,
                   cols = [('orange',None,'orange','Oranges'),
                           ('green',None,'green','Greens'), 
                           ('magenta',None,'magenta','RdPu'),
                           ('purple',None,'purple','Purples'), 
                           ('blue',None,'blue','Blues'),
                           ('k',None,'k','Greys')]):
    '''
    This function plots corner plot of MC chains
    
    Parameters:
    -----------
    - dfs (list): list of pandas dataframe used for the plot
    - output_fig (str): full path to the figure to save
    - ci2d (list): confidence intervals to be considered in 2d
    - ci1d (list): confidence interval to be considered for the histogram
    - truth (list): list of expected value for the parameters
    - truth_style (str): either 'line' or 'star'
    - labels (list): list of label for the datasets
    - gridsize (int): the number of cells in the grid (higher=nicer=slower)
    - smoothing1d (float): the width of the smoothing kernel for 1d distribution
    - smoothing2d (float): the width of the smoothing kernel for 2d distribution
    - linewidth (float): linewidth of the contours
    - alpha (tuple): alpha parameters for the histogram, histogram CI, and contours plot
    - n_levels (int): if set, will draw a 'diffuse' filled contour plot with n_levels
    - zoom (float): controle the axis limits wrt the plotted distribution.
    The give nnumber corresponds to the fractional size of the 2D distribution 
    to add on each side. If negative, will zoom in the plot.
    - add_grid (bool): add the grid in the plot
    - limits (list of tupple): the limit for each parameters
    - figsize (tuple): the size of the figure
    - fontsize (int): the font size
    - cols (list of 4-tupples): deal with the colors for the dataframes. Each tupple
    is for histogram, histogram edges, filled contours, contour edges
    
    Output:
    -------
    - Plots
    '''
    
    # Check type
    if type(dfs) is not list:
        dfs = [dfs]
    
    # Plot length
    Npar = len(dfs[0].columns) # Number of parameters
    Ndat = len(dfs)            # Number of datasets

    # Percentiles
    levels = 1.0-np.array(ci2d)
    levels = np.append(levels, 1.0)
    levels.sort()

    if n_levels is None:
        n_levels = copy.copy(levels)
    
    # Make sure there are enough colors
    icol = 0
    while len(cols) < Ndat:
        cols.append(cols[icol])
        icol = icol+1
    
    # Figure
    plt.figure(figsize=figsize)
    for ip in range(Npar):
        for jp in range(Npar):
            #----- Diagonal histogram
            if ip == jp:
                plt.subplot(Npar,Npar,ip*Npar+jp+1)
                xlims1, xlims2 = [], []
                ylims = []
                # Get the range
                for idx, df in enumerate(dfs, start=0):
                    xlims1.append(np.nanmin(df[df.columns[ip]]))
                    xlims2.append(np.nanmax(df[df.columns[ip]]))
                xmin = np.nanmin(np.array(xlims1))
                xmax = np.nanmax(np.array(xlims2))
                Dx = (xmax - xmin)*zoom
                for idx, df in enumerate(dfs, start=0):
                    if labels is not None:
                        sns.histplot(x=df.columns[ip], data=df, kde=True,
                                     kde_kws={'cut':3, 'bw_adjust':smoothing1d},
                                     color=cols[idx][0], binrange=[xmin-Dx,xmax+Dx],
                                     alpha=alpha[0], edgecolor=cols[idx][1], stat='density', label=labels[idx])
                    else:
                        sns.histplot(x=df.columns[ip], data=df, kde=True,
                                     kde_kws={'cut':3, 'bw_adjust':smoothing1d},
                                     color=cols[idx][0], binrange=[xmin-Dx,xmax+Dx],
                                     alpha=alpha[0], edgecolor=cols[idx][1], stat='density')
                ax = plt.gca()
                ylims.append(ax.get_ylim()[1])
                ax.set_xlim(xmin-Dx, xmax+Dx)
                if limits is not None:
                    ax.set_xlim(limits[ip][0], limits[ip][1])
                ax.set_ylim(0, np.nanmax(np.array(ylims)))

                
                if ci1d is not None:
                    for idx, df in enumerate(dfs, start=0):
                        try:
                            perc = np.percentile(df[df.columns[ip]], [100-(100-ci1d*100)/2.0, (100-ci1d*100)/2.0])
                            # Get the KDE line for filling below
                            xkde = ax.lines[idx].get_xdata()
                            ykde = ax.lines[idx].get_ydata()
                            wkeep = (xkde < perc[0]) * (xkde > perc[1])
                            xkde_itpl = np.append(np.append(perc[1], xkde[wkeep]), perc[0])
                            itpl = interp1d(xkde, ykde)
                            ykde_itpl = itpl(xkde_itpl)
                            perc_max = itpl(perc)
                            
                            ax.vlines(perc[0], 0.0, perc_max[0], linestyle='--', color=cols[idx][0])
                            ax.vlines(perc[1], 0.0, perc_max[1], linestyle='--', color=cols[idx][0])
                            ax.fill_between(xkde_itpl, 0*ykde_itpl, y2=ykde_itpl,
                                            alpha=alpha[1], color=cols[idx][0])
                        except:
                            warnings.warn('could not get and fill bellow KDE line')
                        
                if add_grid:
                    #ax.xaxis.set_major_locator(MultipleLocator((xmax+Dx-(xmin-Dx))/5.0))
                    ax.grid(True, axis='x', linestyle='--')
                else:
                    ax.grid(False)
                        
                if truth is not None:
                    ax.vlines(truth[ip], ax.get_ylim()[0], ax.get_ylim()[1], linestyle=':', color='k')
                    
                plt.yticks([])
                plt.ylabel(None)
                if jp<Npar-1:
                    #plt.xticks([])
                    ax.set_xticklabels([])
                    plt.xlabel(None)
                if ip == 0 and labels is not None:
                    plt.legend(loc='upper left')
                for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] +
                             ax.get_xticklabels() + ax.get_yticklabels()):
                    item.set_fontsize(fontsize)
                    
            #----- Off diagonal 2d plots
            if ip>jp:
                plt.subplot(Npar,Npar,ip*Npar+jp+1)
                xlims1, xlims2 = [], []
                ylims1, ylims2 = [], []
                for idx, df in enumerate(dfs, start=0):
                    xlims1.append(np.nanmin(df[df.columns[jp]]))
                    xlims2.append(np.nanmax(df[df.columns[jp]]))
                    ylims1.append(np.nanmin(df[df.columns[ip]]))
                    ylims2.append(np.nanmax(df[df.columns[ip]]))
                    sns.kdeplot(x=df.columns[jp], y=df.columns[ip], data=df, gridsize=gridsize, 
                                n_levels=n_levels, levels=levels, thresh=levels[0], fill=True, 
                                cmap=cols[idx][3], alpha=alpha[2],
                                bw_adjust=smoothing2d)
                    sns.kdeplot(x=df.columns[jp], y=df.columns[ip], data=df, gridsize=gridsize, 
                                levels=levels[0:-1], color=cols[idx][2], linewidths=linewidth,
                                bw_adjust=smoothing2d)
                ax = plt.gca()
                xmin = np.nanmin(np.array(xlims1))
                xmax = np.nanmax(np.array(xlims2))
                Dx = (xmax - xmin)*zoom
                ymin = np.nanmin(np.array(ylims1))
                ymax = np.nanmax(np.array(ylims2))
                Dy = (ymax - ymin)*zoom

                ax.set_xlim(xmin-Dx, xmax+Dx)
                ax.set_ylim(ymin-Dy, ymax+Dy)
                if limits is not None:
                    ax.set_xlim(limits[jp][0], limits[jp][1])
                    ax.set_ylim(limits[ip][0], limits[ip][1])
                
                if add_grid:
                    #ax.xaxis.set_major_locator(MultipleLocator((xmax+Dx-(xmin-Dx))/5.0))
                    #ax.yaxis.set_major_locator(MultipleLocator((ymax+Dy-(ymin-Dy))/5.0))
                    ax.grid(True, linestyle='--')
                else:
                    ax.grid(False)

                if truth is not None:
                    if truth_style == 'line':
                        ax.vlines(truth[jp], ax.get_ylim()[0], ax.get_ylim()[1], linestyle=':', color='k')
                        ax.hlines(truth[ip], ax.get_xlim()[0], ax.get_xlim()[1], linestyle=':', color='k')
                    if truth_style == 'star':
                        ax.plot(truth[jp], truth[ip], linestyle='', marker="*", color='k', markersize=10)
                    
                if jp > 0:
                    #plt.yticks([])
                    ax.set_yticklabels([])
                    plt.ylabel(None)
                if ip<Npar-1:
                    #ax.set_xticks([])
                    ax.set_xticklabels([])
                    ax.set_xlabel(None)
                for item in ([ax.title, ax.xaxis.label, ax.yaxis.label] +
                             ax.get_xticklabels() + ax.get_yticklabels()):
                    item.set_fontsize(fontsize)
                    
    plt.tight_layout(h_pad=0.0,w_pad=0.0)
    if output_fig is not None:
        plt.savefig(output_fig)


#==================================================
# Plot of the best-fit map
#==================================================

def show_fit_result_ymap(figfile,
                         image, header, noise_mc,
                         model_ymap_sph,
                         mask=None,
                         visu_smooth=10,
                         cmap='Spectral_r',
                         contcol='k'):
    '''
    This function plots the best fit map model
    
    Parameters:
    ----------
    - figfile (str): name of the file to produce
    - image (2d array): the data image
    - header (str): the header
    - noise_mc (3d array): the noise MC realization
    - model_ymap_sph (2d image): the best fit model
    - mask (2d array): the mask if any
    - visu_smooth (float): the smoothing FWHM in arcsec
    - cmap (str): colormap
    - contcol (str): the contour color
    
    Output:
    -------
    - Plots produced
    '''

    #----- Get extra data info
    if mask is None:
        mask = image*0+1

    #----- Compute the rms at given scale
    rms = np.std(gaussian_filter(noise_mc,
                                 sigma=np.array([0,1,1])*visu_smooth/2.35/header['CDELT2']/3600),
                 axis=0)
    levels = [-30,-28,-26,-24,-22,-20,-18,-16,-14,-12,-10,-8,-6,-4,-2,
              2,4,6,8,10,12,14,16,18,20,22,24,26,28,30]

    #----- Compute plot range
    rng_map = [np.amin(mask*gaussian_filter(image,
                                            sigma=visu_smooth/2.35/header['CDELT2']/3600)*1e5),
               np.amax(mask*gaussian_filter(image,
                                            sigma=visu_smooth/2.35/header['CDELT2']/3600)*1e5)]
    
    stdres = np.std((gaussian_filter(image-model_ymap_sph,
                                     sigma=visu_smooth/2.35/header['CDELT2']/3600)*1e5)[mask > 0])
    rng_res = np.array([-1,1]) * stdres * 3
    
    #----- Show the maps
    plt.rcParams.update({'font.size': 12})
    fig = plt.figure(0, figsize=(17, 5))

    # Input data
    ax = plt.subplot(1, 3, 1, projection=WCS(header))
    plt.imshow(mask * gaussian_filter(image,
                                      sigma=visu_smooth/2.35/header['CDELT2']/3600)*1e5,
               cmap=cmap, vmin=rng_map[0], vmax=rng_map[1])
    cb = plt.colorbar()
    plt.contour(mask*gaussian_filter(image,
                                     sigma=visu_smooth/2.35/header['CDELT2']/3600)/rms,
                levels=levels, colors=contcol)
    plt.title('y-Compton data')
    plt.xlabel('R.A. (deg)')
    plt.ylabel('Dec. (deg)')
    
    # Best model
    ax = plt.subplot(1, 3, 2, projection=WCS(header))
    plt.imshow(gaussian_filter(model_ymap_sph,
                               sigma=visu_smooth/2.35/header['CDELT2']/3600)*1e5,
               cmap=cmap, vmin=rng_map[0], vmax=rng_map[1])
    cb = plt.colorbar()
    plt.title('y-Compton best-fit')
    plt.xlabel('R.A. (deg)')
    plt.ylabel(' ')

    # Best model residual
    ax = plt.subplot(1, 3, 3, projection=WCS(header))
    plt.imshow(gaussian_filter(image - model_ymap_sph,
                               sigma=visu_smooth/2.35/header['CDELT2']/3600)*1e5*mask,
               cmap=cmap,
               vmin=rng_res[0], vmax=rng_res[1])
    cb = plt.colorbar()
    plt.contour(mask*gaussian_filter(image-model_ymap_sph,
                                     sigma=visu_smooth/2.35/header['CDELT2']/3600)/rms,
                levels=levels, colors=contcol)
    plt.title('y-Compton residual')
    plt.xlabel('R.A. (deg)')
    plt.ylabel(' ')
    plt.savefig(figfile)
    plt.close()
    

#==================================================
# Plot of the best-fit profile
#==================================================

def show_fit_ycompton_profile(figfile,
                              r2d,
                              data_yprof, data_yprof_err,
                              prof_best,
                              prof_mc,
                              true_compton_profile=None):
    '''
    This function plots the pressure profile results
    
    Parameters:
    ----------
    - figfile (str): name of the file to produce
    - r2d (1d array): the radius in arcmin
    - data_yprof (1d array): the data points
    - data_yprof_err (1d array): the data error
    - prof_best (1d array): the best-fit y compton
    - prof_mc (2d array): the MC resampling of the profile (Nsample x len(r2d))
    - true_y_profile (dict): pass a dictionary containing the profile to compare with
    in the form {'r':array in arcmin, 'p':array in [y]}

    Output:
    -------
    - Plots produced
    '''
    
    #----- Compute model uncertainties
    ci1 = 68.0
    prof_perc1 = np.percentile(prof_mc,
                              (100-(100-ci1)/2.0, 50, (100-ci1)/2.0), axis=0)
    ci2 = 95.0
    prof_perc2 = np.percentile(prof_mc,
                              (100-(100-ci2)/2.0, 50, (100-ci2)/2.0), axis=0)
    
    #----- Plot the result
    plt.rcParams.update({'font.size': 12})
    fig = plt.figure(0, figsize=(7, 5))
    plt.errorbar(r2d, 1e5*data_yprof, 1e5*data_yprof_err, marker='o', ls='', color='k', label='Data')
    plt.plot(r2d, 1e5*prof_best, color='r', ls='-', label='Best-fit model')
    plt.plot(r2d, 1e5*prof_perc1[1,:], color='b', ls='--', label='Median, 68% and 95% C.I.')
    plt.fill_between(r2d, 1e5*prof_perc1[0,:], y2=1e5*prof_perc1[2,:], alpha=0.3, color='blue')
    plt.fill_between(r2d, 1e5*prof_perc2[0,:], y2=1e5*prof_perc2[2,:], alpha=0.15, color='blue')

    if true_compton_profile is not None:
        plt.plot(true_compton_profile['r'], true_compton_profile['y']*1e5, label='True profile', color='orange')
    plt.xlabel('Radius (arcmin)')
    plt.ylabel(r'$10^5 \times y$-Compton')
    plt.xlim(0, np.amax(r2d[np.isfinite(data_yprof)])*1.2)
    plt.legend(fontsize=12)
    plt.savefig(figfile)
    plt.close()
        

#==================================================
# Plot of the best-fit profile
#==================================================

def show_fit_result_pressure_profile(figfile,
                                     r3d,
                                     prof_best,
                                     prof_mc,
                                     true_pressure_profile=None):
    '''
    This function plots the pressure profile results
    
    Parameters:
    ----------
    - figfile (str): name of the file to produce
    - r3d (1d array): the radius in kpc
    - prof_best (1d array): the best-fit pressure in keV/cm^3
    - prof_mc (2d array): the MC resampling of the profile (Nsample x len(r3d))
    - true_pressure_profile (dict): pass a dictionary containing the profile to compare with
    in the form {'r':array in kpc, 'p':array in keV cm-3}
    
    Output:
    -------
    - Plots produced
    '''

    #----- Compute model uncertainties
    ci1 = 68.0
    prof_perc1 = np.percentile(prof_mc,
                              (100-(100-ci1)/2.0, 50, (100-ci1)/2.0), axis=0)
    ci2 = 95.0
    prof_perc2 = np.percentile(prof_mc,
                               (100-(100-ci2)/2.0, 50, (100-ci2)/2.0), axis=0)
    
    #----- Plot the result
    plt.rcParams.update({'font.size': 12})
    fig = plt.figure(0, figsize=(7, 5))
    plt.plot(r3d, prof_best, color='r', ls='-', label='Best-fit model')
    plt.plot(r3d, prof_perc1[1,:], color='b', ls='--', label='Median, 68% and 95% C.I.')
    plt.fill_between(r3d, prof_perc2[0,:], y2=prof_perc2[2,:], alpha=0.15, color='blue')
    plt.fill_between(r3d, prof_perc1[0,:], y2=prof_perc1[2,:], alpha=0.3, color='blue')
    if true_pressure_profile is not None:
        plt.plot(true_pressure_profile['r'], true_pressure_profile['p'], label='True profile', color='orange')
    plt.xlabel('Radius (kpc)')
    plt.ylabel(r'$P_e(r)$ (keV/cm$^3$)')
    plt.xscale('log')
    plt.yscale('log')
    plt.xlim(np.amin(r3d), np.amax(r3d))
    plt.legend(fontsize=12)
    plt.savefig(figfile)
    plt.close()

    
#==================================================
# Plot of contraint on Pk3d
#==================================================

def show_fit_result_pk3d(figfile,
                         k3d,
                         pk3d_best,
                         pk3d_mc,
                         true_pk3d=None):
    '''
    This function plots the deprojected constraint on Pk3d
    given the 2d fit.
    
    Parameters:
    ----------
    - figfile (str): name of the file to produce
    - k3d (1d array): k vector in physical scale
    - pk3d_best (1d array): best fit model for Pk3d
    - pk3d_mc (2d array): Pk3d for parameter sampling
    - true_pk3d (dict): pass a dictionary containing the Pk3d to compare with
    in the form {'k':array in kpc-1, 'pk':array in kpc3}
    
    Output:
    -------
    - Plots produced

    '''
    
    #----- Compute model uncertainties
    ci1 = 68.0
    pk3d_perc1 = np.percentile(pk3d_mc,
                               (100-(100-ci1)/2.0, 50, (100-ci1)/2.0), axis=0)
    ci2 = 95.0
    pk3d_perc2 = np.percentile(pk3d_mc,
                               (100-(100-ci2)/2.0, 50, (100-ci2)/2.0), axis=0)

    #----- Plot the result
    plt.rcParams.update({'font.size': 12})
    fig = plt.figure(0, figsize=(7, 5))
    plt.loglog(k3d, np.sqrt(4*np.pi*k3d**3*pk3d_best), color='r', ls='-', label='Best-fit')
    plt.loglog(k3d, np.sqrt(4*np.pi*k3d**3*pk3d_perc1[1,:]), color='b', ls='--', label='Median, 68% and 95% C.I.')

    plt.fill_between(k3d, np.sqrt(4*np.pi*k3d**3*pk3d_perc1[0,:]), np.sqrt(4*np.pi*k3d**3*pk3d_perc1[2,:]),
                     color='blue', alpha=0.3)
    plt.fill_between(k3d, np.sqrt(4*np.pi*k3d**3*pk3d_perc2[0,:]), np.sqrt(4*np.pi*k3d**3*pk3d_perc2[2,:]),
                     color='blue', alpha=0.15)
    
    if true_pk3d is not None:
        plt.plot(true_pk3d['k'], np.sqrt(4*np.pi*true_pk3d['k']**3*true_pk3d['pk']),
                 label='True $P_k$', color='orange')
    plt.xlabel(r'$k$ (kpc$^{-1}$)')
    plt.ylabel(r'$\sqrt{4 \pi k^3 P(k)}$')
    plt.ylim(np.amax(np.sqrt(4*np.pi*k3d**3*pk3d_perc1[1,:]))*1e-3,
             np.amax(np.sqrt(4*np.pi*k3d**3*pk3d_perc1[1,:]))*5)
    plt.xlim(np.amin(k3d), np.amax(k3d))
    plt.legend(fontsize=12)
    plt.savefig(figfile)
    plt.close()


#==================================================
# Plot of contraint on Pk2d
#==================================================

def show_fit_result_pk2d(figfile,
                         k2d,
                         pk2d_data,
                         pk2d_modref,
                         pk2d_data_err_model,
                         pk2d_data_err_noise,
                         pk2d_data_err_bkg,
                         pk2d_mc,
                         pk2d_noise_best, pk2d_noise_mc,                         
                         pk2d_bkg_best, pk2d_bkg_mc,                         
                         true_pk2d=None):
    '''
    This function plots the deprojected constraint on Pk3d
    given the 2d fit.
    
    Parameters:
    ----------
    - figfile (str): name of the file to produce
    - k2d (1d array): k vector in physical scale
    - pk2d_data (1d array): data points
    - pk2d_modref (1d array): reference model used for uncertainty computation
    - pk2d_data_err_model (1d array): uncertainty acssociated with the model
    - pk2d_data_err_noise (1d array): uncertainty associated with the noise
    - pk2d_data_err_bkg (1d array): uncertainty associated with the bkg
    - pk2d_mc (2d array): MC ressampling for the cluster model contribution
    - pk2d_noise_best (1d array): bestfit noise contribution
    - pk2d_noise_mc (2d array): MC ressampling for the noise contribution
    - pk2d_bkg_best (1d array): bestfit bkg contribution
    - pk2d_bkg_mc (2d array): MC ressampling for the bkg contribution
    - true_pk2d (dict): pass a dictionary containing the Pk3d to compare with
    in the form {'k':array in arcsec-1, 'pk':array in arcsec2}
    
    Output:
    -------
    - Plots produced

    '''
    
    #----- Compute model uncertainties
    ci1 = 68.0
    pk2d_perc1 = np.percentile(pk2d_mc,
                              (100-(100-ci1)/2.0, 50, (100-ci1)/2.0), axis=0)
    pk2d_noise_perc1 = np.percentile(pk2d_noise_mc,
                                    (100-(100-ci1)/2.0, 50, (100-ci1)/2.0), axis=0)
    pk2d_bkg_perc1 = np.percentile(pk2d_bkg_mc,
                                   (100-(100-ci1)/2.0, 50, (100-ci1)/2.0), axis=0)

    ci2 = 95.0
    pk2d_perc2 = np.percentile(pk2d_mc,
                              (100-(100-ci2)/2.0, 50, (100-ci2)/2.0), axis=0)
    pk2d_noise_perc2 = np.percentile(pk2d_noise_mc,
                                    (100-(100-ci2)/2.0, 50, (100-ci2)/2.0), axis=0)
    pk2d_bkg_perc2 = np.percentile(pk2d_bkg_mc,
                                    (100-(100-ci2)/2.0, 50, (100-ci2)/2.0), axis=0)

    #----- Compute uncertainty
    err_noise = np.sqrt(2*np.pi*k2d**2)/(2*np.sqrt(pk2d_data))*pk2d_data_err_noise
    err_nbkg  = np.sqrt(2*np.pi*k2d**2)/(2*np.sqrt(pk2d_data))*(pk2d_data_err_noise**2
                                                                + pk2d_data_err_bkg**2)**0.5
    err_tot   = np.sqrt(2*np.pi*k2d**2)/(2*np.sqrt(pk2d_data))*(pk2d_data_err_model**2
                                                                + pk2d_data_err_noise**2
                                                                + pk2d_data_err_bkg**2)**0.5
    
    #----- Plot the result
    plt.rcParams.update({'font.size': 12})
    fig = plt.figure(0, figsize=(7, 5))
    frame1 = fig.add_axes((.15,.3,.8,.6))

    # Data
    plt.errorbar(k2d, np.sqrt(2*np.pi*k2d**2*pk2d_data), err_tot, marker='.', ls='', color='lightgrey',
                 label='Data (noise+bkg+model uncertainties)')
    plt.errorbar(k2d, np.sqrt(2*np.pi*k2d**2*pk2d_data), err_nbkg, marker='.', ls='', color='grey',
                 label='Data (noise+bkg uncertainty)')
    plt.errorbar(k2d, np.sqrt(2*np.pi*k2d**2*pk2d_data), err_noise, marker='.', ls='', color='k',
                 label='Data (noise uncertainty)')

    # Model reference
    plt.plot(k2d, np.sqrt(2*np.pi*k2d**2*(pk2d_modref)), color='magenta', label='Best-fit model and intrinsic scatter')
    up, low = pk2d_modref+pk2d_data_err_model, pk2d_modref-pk2d_data_err_model
    low[low<=0] = 0
    plt.fill_between(k2d, np.sqrt(2*np.pi*k2d**2*(up)), y2=np.sqrt(2*np.pi*k2d**2*(low)),
                     color='magenta', alpha=0.3)
    
    # Model fit
    plt.loglog(k2d, np.sqrt(2*np.pi*k2d**2*pk2d_perc1[1,:]), color='b', ls='--', label='Model median, 68% and 95% C.I.')
    plt.fill_between(k2d, np.sqrt(2*np.pi*k2d**2*pk2d_perc1[0,:]), np.sqrt(2*np.pi*k2d**2*pk2d_perc1[2,:]),
                     color='blue', alpha=0.3)
    plt.fill_between(k2d, np.sqrt(2*np.pi*k2d**2*pk2d_perc2[0,:]), np.sqrt(2*np.pi*k2d**2*pk2d_perc2[2,:]),
                     color='blue', alpha=0.15)
    
    # Noise fit
    plt.loglog(k2d, np.sqrt(2*np.pi*k2d**2*pk2d_noise_perc1[1,:]), color='darkcyan', ls='--',
               label='Noise median, 68% and 95% C.I.')
    plt.fill_between(k2d, np.sqrt(2*np.pi*k2d**2*pk2d_noise_perc1[0,:]),
                     np.sqrt(2*np.pi*k2d**2*pk2d_noise_perc1[2,:]),
                     color='darkcyan', alpha=0.3)
    plt.fill_between(k2d, np.sqrt(2*np.pi*k2d**2*pk2d_noise_perc2[0,:]),
                     np.sqrt(2*np.pi*k2d**2*pk2d_noise_perc2[2,:]),
                     color='darkcyan', alpha=0.15)

    # Bkg fit
    plt.loglog(k2d, np.sqrt(2*np.pi*k2d**2*pk2d_bkg_perc1[1,:]), color='darkorange', ls='--',
               label='Bkg median, 68% and 95% C.I.')
    plt.fill_between(k2d, np.sqrt(2*np.pi*k2d**2*pk2d_bkg_perc1[0,:]),
                     np.sqrt(2*np.pi*k2d**2*pk2d_bkg_perc1[2,:]),
                     color='darkorange', alpha=0.3)
    plt.fill_between(k2d, np.sqrt(2*np.pi*k2d**2*pk2d_bkg_perc2[0,:]),
                     np.sqrt(2*np.pi*k2d**2*pk2d_bkg_perc2[2,:]),
                     color='darkorange', alpha=0.15)

    # Total
    plt.loglog(k2d, np.sqrt(2*np.pi*k2d**2*(pk2d_noise_best + pk2d_bkg_best + pk2d_modref)),
               color='red', ls='-', label='Total best-fit')
        
    if true_pk2d is not None:
        plt.plot(true_pk2d['k'], np.sqrt(2*np.pi*true_pk2d['k']**2*true_pk2d['pk']),
                 label='True $P_k$', color='orange')
    plt.ylabel(r'$\sqrt{2 \pi k^2 P(k)}$')
    plt.xscale('log')
    plt.yscale('log')
    plt.xlim(np.amin(k2d)*0.9, np.amax(k2d)*1.1)
    
    plt.ylim(np.nanmax(np.sqrt(2*np.pi*k2d**2*pk2d_data)+err_tot)*1e-3,
             np.nanmax(np.sqrt(2*np.pi*k2d**2*pk2d_data)+err_tot)*2)
    
    plt.legend(fontsize=10)

    frame2 = fig.add_axes((.15,.1,.8,.2))
    dat = np.sqrt(2*np.pi*k2d**2*pk2d_data)
    mod = np.sqrt(2*np.pi*k2d**2*(pk2d_noise_best + pk2d_bkg_best + pk2d_modref))
    plt.plot(k2d, (dat-mod)/err_tot, marker='o', color='lightgrey')
    plt.plot(k2d, (dat-mod)/err_nbkg, marker='o', color='grey')
    plt.plot(k2d, (dat-mod)/err_noise, marker='o', color='k')
    plt.axhline(0, color='k', linestyle='-')
    plt.axhline(+3, color='k', linestyle='--')
    plt.axhline(-3, color='k', linestyle='--')
    plt.axhline(+5, color='k', linestyle='--')
    plt.axhline(-5, color='k', linestyle='--')
    plt.xscale('log')
    plt.xlabel(r'$k$ (kpc$^{-1}$)')
    plt.ylabel('Residual ($\sigma$)')
    
    plt.savefig(figfile)
    plt.close()


#==================================================
# Plot of delta_y / y fitted
#==================================================

def show_input_delta_ymap(figfile,
                          image,
                          data_image,
                          model_ymap_sph1,
                          model_ymap_sph2,
                          w8,
                          header,
                          noise_mc,
                          mask=None,
                          visu_smooth=10,
                          cmap='Spectral_r',
                          contcol='k'):
    '''
    This function plots the best fit map model
    
    Parameters:
    ----------
    - figfile (str): name of the file to produce
    - image (2d array): the data image
    - header (str): the header
    - model_ymap_sph1 (2d image): the best fit model
    - model_ymap_sph2 (2d image): the best fit model deconvolved
    - w8 (2d array): the weight map
    - mask (2d array): the mask if any
    - visu_smooth (float): the smoothing FWHM in arcsec
    - cmap (str): colormap
    - contcol (str): contour color
    
    Output:
    -------
    - Plots produced
    '''

    #----- Get extra data info
    if mask is None:
        mask = image*0+1

    #----- Compute the rms at given scale
    rms = np.std(gaussian_filter(noise_mc,
                                 sigma=np.array([0,1,1])*visu_smooth/2.35/header['CDELT2']/3600),
                 axis=0)
    rms_img = np.std(gaussian_filter(noise_mc*w8[np.newaxis,:,:]/model_ymap_sph2,
                                     sigma=np.array([0,1,1])*visu_smooth/2.35/header['CDELT2']/3600),
                 axis=0)
    rms_img[rms_img <=0] = np.nan
    levels = [-30,-28,-26,-24,-22,-20,-18,-16,-14,-12,-10,-8,-6,-4,-2,
              2,4,6,8,10,12,14,16,18,20,22,24,26,28,30]

    #---------- Plot
    plt.rcParams.update({'font.size': 12})
    fig = plt.figure(0, figsize=(23, 5))

    # Input data
    ax = plt.subplot(1, 4, 1, projection=WCS(header))
    plt.imshow(mask * gaussian_filter(image,
                                      sigma=visu_smooth/2.35/header['CDELT2']/3600)*1e5,cmap=cmap)
    cb = plt.colorbar()
    plt.contour(mask*gaussian_filter(image,
                                     sigma=visu_smooth/2.35/header['CDELT2']/3600)/rms,
                levels=levels, colors=contcol)
    plt.title(r'y-Compton data ($\times 10^5$)')
    plt.xlabel('R.A. (deg)')
    plt.ylabel('Dec. (deg)')
    
    # Best model
    ax = plt.subplot(1, 4, 2, projection=WCS(header))
    plt.imshow(model_ymap_sph1*1e5,cmap=cmap)
    cb = plt.colorbar()
    plt.title(r'y-Compton best-fit ($\times 10^5$)')
    plt.xlabel('R.A. (deg)')
    plt.ylabel(' ')

    # w8
    ax = plt.subplot(1, 4, 3, projection=WCS(header))
    plt.imshow(w8,cmap=cmap)
    cb = plt.colorbar()
    plt.title('Weight')
    plt.xlabel('R.A. (deg)')
    plt.ylabel(' ')
    
    # dy/y w8
    ax = plt.subplot(1, 4, 4, projection=WCS(header))
    plt.imshow(gaussian_filter(data_image,
                               sigma=visu_smooth/2.35/header['CDELT2']/3600),cmap=cmap)
    cb = plt.colorbar()
    plt.contour(gaussian_filter(data_image,
                                sigma=visu_smooth/2.35/header['CDELT2']/3600)/rms_img,
                levels=levels, colors=contcol)
    
    plt.title(r'$\Delta y / y \times W$')
    plt.xlabel('R.A. (deg)')
    plt.ylabel(' ')
    
    plt.savefig(figfile)
    plt.close()


#==================================================
# Plot of covariance
#==================================================

def show_fit_result_covariance(figfile,
                               covmat_data,
                               covmat_bfmodel,
                               covmat_model=None,
                               covmat_bkg=None):
    '''
    This function plots the best fit map model
    
    Parameters:
    ----------
    - figfile (str): name of the file to produce
    - covmat_data (2d array): bin to bin noise covariance matrix
    - covmat_bfmodel (2d array): bin to bin best fit model covariance matrix
    - covmat_model (2d array): bin to bin reference model covariance matrix
    - covmat_nkg (2d array): bin to bin background covariance matrix

    Output:
    -------
    - Plots produced
    '''
    
    Nrow = 2
    if covmat_model is not None: Nrow += 1
    if covmat_bkg   is not None: Nrow += 1
    
    plt.rcParams.update({'font.size': 12})
    fig = plt.figure(0, figsize=(12, 5*Nrow))

    # Covariance data
    ax = plt.subplot(Nrow, 2, 1)
    plt.imshow(covmat_data)
    cb = plt.colorbar()
    plt.title(r'Noise covariance matrix')
    plt.xlabel('k bin')
    plt.ylabel('k bin')
    
    # Correlation data
    ax = plt.subplot(Nrow, 2, 2)
    plt.imshow(utils.correlation_from_covariance(covmat_data))
    cb = plt.colorbar()
    plt.title(r'Noise correlation matrix')
    plt.xlabel('k bin')
    plt.ylabel('k bin')

    # Covariance best fit model
    ax = plt.subplot(Nrow, 2, 3)
    plt.imshow(covmat_bfmodel)
    cb = plt.colorbar()
    plt.title(r'Best fit model covariance matrix')
    plt.xlabel('k bin')
    plt.ylabel('k bin')
    
    # Correlation best fit model
    ax = plt.subplot(Nrow, 2, 4)
    plt.imshow(utils.correlation_from_covariance(covmat_bfmodel))
    cb = plt.colorbar()
    plt.title(r'Best fit model correlation matrix')
    plt.xlabel('k bin')
    plt.ylabel('k bin')

    idx = 0
    if covmat_model is not None:
        idx += 2
        # Covariance model
        ax = plt.subplot(Nrow, 2, 5)
        plt.imshow(covmat_model)
        cb = plt.colorbar()
        plt.title(r'Ref model covariance matrix')
        plt.xlabel('k bin')
        plt.ylabel('k bin')
        
        # Correlation model
        ax = plt.subplot(Nrow, 2, 6)
        plt.imshow(utils.correlation_from_covariance(covmat_model))
        cb = plt.colorbar()
        plt.title(r'Ref model correlation matrix')
        plt.xlabel('k bin')
        plt.ylabel('k bin')

    if covmat_bkg is not None:
    
        # Covariance model
        ax = plt.subplot(Nrow, 2, 5+idx)
        plt.imshow(covmat_bkg)
        cb = plt.colorbar()
        plt.title(r'Background covariance matrix')
        plt.xlabel('k bin')
        plt.ylabel('k bin')
        
        # Correlation model
        ax = plt.subplot(Nrow, 2, 6+idx)
        plt.imshow(utils.correlation_from_covariance(covmat_bkg))
        cb = plt.colorbar()
        plt.title(r'Background correlation matrix')
        plt.xlabel('k bin')
        plt.ylabel('k bin')
    
    plt.savefig(figfile)
    plt.close()



"""
This file contain a subclass of the main_model.py module and ClusterModel class. It
is dedicated to the computing of mock observables.

"""

#==================================================
# Requested imports
#==================================================

import numpy as np
from scipy.interpolate import interp1d
from astropy.coordinates import SkyOffsetFrame, SkyCoord
import astropy.units as u
import astropy.constants as cst
from scipy.spatial.transform import Rotation as R

from minot.ClusterTools import map_tools

from pitszi import utils


#==================================================
# Mock class
#==================================================

class ModelSampling(object):
    """ ModelSampling class
    This class serves as a parser to the main Model class, to
    include the subclass ModelSampling in this other file.

    Attributes
    ----------
    The attributes are the same as the Model class, see model_main.py

    Methods
    ----------
    - get_map_header
    - get_map_reso
    - get_map_fov
    - get_map_center
    - get_3dgrid
    
    """

    #==================================================
    # Extract the header
    #==================================================
    
    def get_map_header(self):
        """
        Extract the header of the map
        
        Parameters
        ----------

        Outputs
        ----------
        - header (astropy object): the header associated to the map

        """

        # Get the needed parameters in case of map header
        if self._map_header is not None:
            header = self._map_header
            
        # Get the needed parameters in case of set-by-hand map parameters
        elif (self._map_center is not None) and (self._map_reso is not None) and (self._map_fov is not None):
            header = map_tools.define_std_header(self._map_center.icrs.ra.to_value('deg'),
                                                 self._map_center.icrs.dec.to_value('deg'),
                                                 self._map_fov.to_value('deg')[0],
                                                 self._map_fov.to_value('deg')[1],
                                                 self._map_reso.to_value('deg'))
        
        # Otherwise there is a problem
        else:
            raise TypeError("A header, or the map_center & map_reso & map_fov should be defined.")

        return header        

    
    #==================================================
    # Extract the resolution
    #==================================================
    
    def get_map_reso(self, physical=False):
        """
        Extract the resolution of the map
        
        Parameters
        ----------
        - physical (bool): set to true to have the size in kpc

        Outputs
        ----------
        - reso (quantity): the resolution associated to the map

        """

        header = self.get_map_header()

        dx =  np.abs(header['CDELT1'])
        dy =  np.abs(header['CDELT2'])
        if dx != dy:
            raise ValueError('The map should have the same pixel resolution along x and y')

        if physical:
            map_reso = (dx*u.deg).to_value('radian')*self._D_ang.to('kpc')
        else:
            map_reso = dx*u.deg  
        
        return map_reso
        

    #==================================================
    # Extract the field of view
    #==================================================
    
    def get_map_fov(self, physical=False):
        """
        Extract the field of view of the map
        
        Parameters
        ----------
        - physical (bool): set to true to have the size in kpc

        Outputs
        ----------
        - fov (quantity): the field of view associated to the map

        """

        header = self.get_map_header()
        reso   = self.get_map_reso().to_value('deg')
        Nx     = header['NAXIS1']
        Ny     = header['NAXIS2']

        if physical:
            map_fov = [Nx*reso*np.pi/180*self._D_ang.to_value('kpc'),
                       Ny*reso*np.pi/180*self._D_ang.to_value('kpc')]*u.kpc
        else:
            map_fov = [Nx*reso, Ny*reso]*u.deg
        
        return map_fov

    
    #==================================================
    # Get the center of the map
    #==================================================
    
    def get_map_center(self):
        """
        Extract the coordinates of the center of the map
        
        Parameters
        ----------
        
        Outputs
        ----------
        - map_center (SkyCoord object): the center of the map
        
        """
        
        header = self.get_map_header()
        Nx = header['NAXIS1']
        Ny = header['NAXIS2']
        ramap, decmap = map_tools.get_radec_map(header) # indexing is xy in this function

        #----- Case of even even
        if (Nx/2.0 == int(Nx/2.0)) and (Ny/2.0 == int(Ny/2.0)):
            ramap_center  = np.mean(ramap[int(Ny/2.0)-1:int(Ny/2.0), int(Nx/2.0)-1:int(Nx/2.0)])
            decmap_center = np.mean(decmap[int(Ny/2.0)-1:int(Ny/2.0), int(Nx/2.0)-1:int(Nx/2.0)])
     
        #----- Case of odd odd
        if (Nx/2.0 != int(Nx/2.0)) and (Ny/2.0 != int(Ny/2.0)):
            ramap_center  = ramap[int(Ny/2.0), int(Nx/2.0)]
            decmap_center = decmap[int(Ny/2.0), int(Nx/2.0)]
     
        #----- Case of even odd
        if (Nx/2.0 == int(Nx/2.0)) and (Ny/2.0 != int(Ny/2.0)):
            ramap_center  = np.mean(ramap[int(Ny/2.0), int(Nx/2.0)-1:int(Nx/2.0)])
            decmap_center = np.mean(decmap[int(Ny/2.0), int(Nx/2.0)-1:int(Nx/2.0)])
     
        #----- Case of odd even
        if (Nx/2.0 != int(Nx/2.0)) and (Ny/2.0 == int(Ny/2.0)):
            ramap_center  = np.mean(ramap[int(Ny/2.0)-1:int(Ny/2.0), int(Nx/2.0)])
            decmap_center = np.mean(decmap[int(Ny/2.0)-1:int(Ny/2.0), int(Nx/2.0)])
                
        map_center = SkyCoord(ramap_center*u.deg, decmap_center*u.deg, frame='icrs')
     
        return map_center        

    
    #==================================================
    # Build the 3d grid
    #==================================================
    
    def get_3dgrid(self):
        """
        Build the physical 3D grid.
        
        Parameters
        ----------
        
        Outputs
        ----------
        - Nx,Ny,Nz (int): the grid number of pixel along each axis
        - proj_reso (float): the pixel size in the projected direction in kpc
        - los_reso (float): the pixel size alonog the line of sight in kpc
        
        """

        # Get the X,Y direction
        header = self.get_map_header()                       # header of the output map
        Nx = header['NAXIS1']
        Ny = header['NAXIS2']
        if np.abs(header['CDELT1']) != np.abs(header['CDELT2']):
            raise ValueError("Only maps with the same resolution towards x and y are allowed.")
        proj_reso_deg = np.abs(header['CDELT2'])
        proj_reso = self._D_ang.to_value('kpc')*proj_reso_deg*np.pi/180
        
        # Define the z axis
        Nz_min = (self._los_size / self._los_reso).to_value('')
        Nz = int(np.ceil(Nz_min))
        
        return Nx, Ny, Nz, proj_reso, proj_reso, self._los_reso.to_value('kpc')


    #==================================================
    # Define the maximum k for having isotropy in 3D
    #==================================================
    
    def get_kmax_isotropic(self, physical=True):
        """
        Compute the maximum value of k so that we have isotropic sampling.
        I.e., all k values beyond min (kmax_x, kmax_y, kmax_z) are not 
        isotropic.
        
        Parameters
        ----------
        - Physical (bool): set to true to have kpc units, otherwise it will be arcsec

        Outputs
        ----------
        - kmax (float): the isotropic kmax in unit of kpc^-1
        """

        Nx, Ny, Nz, proj_reso, proj_reso, los_reso = self.get_3dgrid()
        
        # check if even or odd and get kmax along each dimension
        if (Nx % 2) == 0:
            kmax_x = 1/(2*proj_reso)
        else:
            kmax_x = (Nx-1)/(2*Nx*proj_reso)
    
        if (Ny % 2) == 0: 
            kmax_y = 1/(2*proj_reso)
        else:
            kmax_y = (Ny-1)/(2*Ny*proj_reso)
    
        if (Nz % 2) == 0: 
            kmax_z = 1/(2*los_reso)
        else:
            kmax_z = (Nz-1)/(2*Nz*los_reso)
            
        # Take the min of the kmax along each axis as kmax isotropic
        kmax_iso = np.amin([kmax_x, kmax_y, kmax_z])

        if physical:
            kmax_iso = kmax_iso*u.kpc**-1
        else:
            kmax_iso = (kmax_iso*self._D_ang.to_value('kpc')*u.rad**-1).to_value('arcsec-1')
            
        return kmax_iso

    
    #==================================================
    # Give information relative to sampling
    #==================================================
    
    def give_sampling_information(self):
        """
        Output information relative to the sampling.
        
        Parameters
        ----------
        
        Outputs
        ----------
        Print information
        
        """

        Nx, Ny, Nz, proj_reso, proj_reso, los_reso = self.get_3dgrid()
        
        print('=====================================================')
        print('===== Information relative to the grid sampling =====')                
        print('=====================================================')
        print('   Grid size :', Nx, ',', Ny, ',', Nz)
        reso_kpc = self.get_map_reso(physical=True).to_value('kpc')
        reso_arcsec = self.get_map_reso(physical=False).to_value('arcsec')
        reso_kpc_s = "{:^10.1f}".format(reso_kpc)
        reso_arcsec_s = "{:^10.1f}".format(reso_arcsec)
        print('   Pixel size :  ', reso_kpc_s, ' kpc ; ', reso_arcsec_s, ' arcsec')

        fov_kpc = self.get_map_fov(physical=True).to_value('kpc')
        fov_arcsec = self.get_map_fov(physical=False).to_value('arcmin')
        print('   Fov size :  [', "{:^10.1f}".format(fov_kpc[0]),
              ',', "{:^10.1f}".format(fov_kpc[1]),
              '] kpc ; [',
              "{:^10.3f}".format(fov_arcsec[0]),',',
              "{:^10.3f}".format(fov_arcsec[1]),'] arcmin')
        print('   L.o.S. resolution :     ', "{:^10.1f}".format(los_reso),' kpc')
        print('   Map center :  ', self.get_map_center())
        k_proj = np.fft.fftfreq(Nx, reso_arcsec)
        print('   k min/max projected :     ',
              "{:^10.6f}".format(np.amin(k_proj[k_proj>0])),'/',
              "{:^10.6f}".format(np.amax(k_proj)),' 1/arcsec')
        k_proj = np.fft.fftfreq(Nx, proj_reso)
        print('   k min/max projected :     ',
              "{:^10.6f}".format(np.amin(k_proj[k_proj>0])),'/',
              "{:^10.6f}".format(np.amax(k_proj)), ' 1/kpc')
        k_z = np.fft.fftfreq(Nz, los_reso)
        conv = self.D_ang.to_value('kpc')/(1*u.rad).to_value('arcsec')
        print('   k min/max L.o.S. (eq.) :  ',
              "{:^10.6f}".format(conv*np.amin(k_z[k_z>0])),'/',
              "{:^10.6f}".format(conv*np.amax(k_z)), ' 1/arcsec')
        print('   k min/max L.o.S. :        ',
              "{:^10.6f}".format(np.amin(k_z[k_z>0])),'/',
              "{:^10.6f}".format(np.amax(k_z)), ' 1/kpc')
        kmax_iso = self.get_kmax_isotropic().to_value('kpc-1')
        print('   k max Nyquist :           ',
              "{:^10.6f}".format(conv*kmax_iso), ' 1/arcsec')
        print('   k max Nyquist :           ',
              "{:^10.6f}".format(kmax_iso), ' 1/kpc')
        print('=====================================================')


        



"""
This file contains generic utilities for various making calculations used in pitszi

"""

import numpy as np


#==================================================
# Check array
#==================================================

def check_qarray(qarr, unit=None):
    """
    Make sure quantity array are arrays

    Parameters
    ----------
    - qarr (quantity): array or float, homogeneous to some unit

    Outputs
    ----------
    - qarr (quantity): quantity array

    """

    if unit is None:
        if type(qarr) == float or type(qarr) == np.float64:
            qarr = np.array([qarr])
            
    else:
        try:
            test = qarr.to(unit)
        except:
            raise TypeError("Unvalid unit for qarr")

        if type(qarr.to_value()) == float or type(qarr.to_value()) == np.float64:
            qarr = np.array([qarr.to_value()]) * qarr.unit

    return qarr


#==================================================
# Def array based on point per decade, min and max
#==================================================

def sampling_array(xmin, xmax, NptPd=10, unit=False):
    """
    Make an array with a given number of point per decade
    from xmin to xmax

    Parameters
    ----------
    - xmin (quantity): min value of array
    - xmax (quantity): max value of array
    - NptPd (int): the number of point per decade

    Outputs
    ----------
    - array (quantity): the array

    """

    if unit:
        my_unit = xmin.unit
        array = np.logspace(np.log10(xmin.to_value(my_unit)),
                            np.log10(xmax.to_value(my_unit)),
                            int(NptPd*(np.log10(xmax.to_value(my_unit)/xmin.to_value(my_unit)))))*my_unit
    else:
        array = np.logspace(np.log10(xmin), np.log10(xmax), int(NptPd*(np.log10(xmax/xmin))))

    return array


#==================================================
# Integration loglog space with trapezoidale rule
#==================================================

def trapz_loglog(y, x, axis=-1, intervals=False):
    """
    Integrate along the given axis using the composite trapezoidal rule in
    loglog space. Integrate y(x) along given axis in loglog space. y can be a function
    with multiple dimension. This follows the script in the Naima package.
    
    Parameters
    ----------
    - y (array_like): Input array to integrate.
    - x (array_like):  optional. Independent variable to integrate over.
    - axis (int): Specify the axis.
    - intervals (bool): Return array of shape x not the total integral, default: False
    
    Returns
    -------
    - trapz (float): Definite integral as approximated by trapezoidal rule in loglog space.
    """
    
    log10 = np.log10
    
    #----- Check for units
    try:
        y_unit = y.unit
        y = y.value
    except AttributeError:
        y_unit = 1.0
    try:
        x_unit = x.unit
        x = x.value
    except AttributeError:
        x_unit = 1.0

    y = np.asanyarray(y)
    x = np.asanyarray(x)

    #----- Define the slices
    slice1 = [slice(None)] * y.ndim
    slice2 = [slice(None)] * y.ndim
    slice1[axis] = slice(None, -1)
    slice2[axis] = slice(1, None)
    slice1, slice2 = tuple(slice1), tuple(slice2)

    #----- arrays with uncertainties contain objects, remove tiny elements
    if y.dtype == "O":
        from uncertainties.unumpy import log10
        # uncertainties.unumpy.log10 can't deal with tiny values see
        # https://github.com/gammapy/gammapy/issues/687, so we filter out the values
        # here. As the values are so small it doesn't affect the final result.
        # the sqrt is taken to create a margin, because of the later division
        # y[slice2] / y[slice1]
        valid = y > np.sqrt(np.finfo(float).tiny)
        x, y = x[valid], y[valid]

    #----- reshaping x
    if x.ndim == 1:
        shape = [1] * y.ndim
        shape[axis] = x.shape[0]
        x = x.reshape(shape)
        
    #-----
    with np.errstate(invalid="ignore", divide="ignore"):
        # Compute the power law indices in each integration bin
        b = log10(y[slice2] / y[slice1]) / log10(x[slice2] / x[slice1])
        
        # if local powerlaw index is -1, use \int 1/x = log(x); otherwise use normal
        # powerlaw integration
        trapzs = np.where(np.abs(b + 1.0) > 1e-10,
                          (y[slice1] * (x[slice2] * (x[slice2] / x[slice1]) ** b - x[slice1]))
                          / (b + 1),
                          x[slice1] * y[slice1] * np.log(x[slice2] / x[slice1]))
        
    tozero = (y[slice1] == 0.0) + (y[slice2] == 0.0) + (x[slice1] == x[slice2])
    trapzs[tozero] = 0.0
    
    if intervals:
        return trapzs * x_unit * y_unit
    
    ret = np.add.reduce(trapzs, axis) * x_unit * y_unit
    
    return ret


#==================================================
# Extract correlation matrix from covarariance
#==================================================

def correlation_from_covariance(covariance):
    """
    Compute the correlation matrix from the covariance
    
    Parameters
    ----------
    - covariance (2d array): the covariance matrix

    Output
    ------
    - correlation (2d array): the correlation matrix

    """
    
    v = np.sqrt(np.diag(covariance))
    outer_v = np.outer(v, v)
    correlation = covariance / outer_v
    correlation[covariance == 0] = 0
    
    return correlation


#==================================================
# Extract covariance matrix from correlation + std
#==================================================

def covariance_from_correlation(correlation_matrix, std):
    """
    Convert correlation matrix to covariance matrix.
    
    Parameters
    ----------
    - correlation_matrix (2d array): numpy array, the correlation matrix
    - std (1d array): numpy array, standard dev along the diagonal of the covariance matrix
    
    Returns:
    - cov_matrix (2d array): numpy array, the covariance matrix
    """

    cov_matrix = correlation_matrix * np.outer(std, std)
    
    return cov_matrix




"""
This file contains the Data class. It is dedicated to the construction of a
Data object, definined by the observational properties of the data.
"""

#==================================================
# Requested imports
#==================================================

import os
import numpy as np
import astropy.units as u
from astropy.io import fits
from astropy.coordinates import SkyCoord
from astropy import constants as const
from astropy.wcs import WCS
from scipy.interpolate import interp1d
from scipy.ndimage import gaussian_filter
import copy
import pickle
import pprint
import dill

from minot.ClusterTools import map_tools
from pitszi import utils_pk
from pitszi import title


#==================================================
# Data class
#==================================================

class Data():
    """ Data class.
    This class defines a data object. 

    Attributes
    ----------
    - image (2d array): the data image
    - header (str): the header associated with image
    - mask (2d array): the mask associated with the image data
    - psf_fwhm (quantity): the FWHM of the PSF as a quantity homogeneous to arcsec
    - transfer_function (dict): the transfer function as {'k':k*u.arcsec, 'TF':TF}
    - noise_rms (2d np.array): rms associated with the image
    - noise_covmat (2d np.array): noise covariance matrix associated with flattened image
    - noise_model (list of 2 element): 
        b) returns the noise amplitude as a function of position. This can be a map, or a 
           function that depends on radius from the center in arcsec.
        a) function that give the noise power spectrum, once normalized by its amplitude,
           as a function of k in arcsec^-1
    - noise_mc (3d np array): Monte Carlo noise realizations. Shape = (Nmc, Ny, Nx)
    - silent (bool): set to True to give information
    - output_dir (str): directory where saving outputs

    Methods
    ----------
    - print_param: print the current Data parameters
    - save_data: save the data object
    - load_data: load a pre-existing data object
    - set_nika2_reference_noise_model: set the noise model to a baseline
    - set_nika2_reference_tf: set the transfer fucntion to a NIKA2 reference model
    - get_noise_monte_carlo_from_model: compute noise monte carlo given the noise model
    - get_noise_monte_carlo_from_covariance: compute noise MC realization from the covariance matrix
    - get_noise_covariance_from_model: compute the covariance given a model
    - get_noise_covariance_from_noise_mc: compute the covariance given a input noise MC
    - save_noise_covariance: save the covariance matrix
    - load_noise_covariance: load a pre-existing covariance matrix
    - set_noise_model_from_jackknife: use a jackknife to compute fir the noise model
    - set_noise_model_from_mc: use noise MC to define a noise model
    - set_noise_model_from_covariance: use noise covariance matrix to define a noise model
    - get_noise_rms_from_mc: get the noise rms given the noise MC realizations
    - get_noise_rms_from_covariance: get the noise rms given the noise covariance matrix
    - get_noise_rms_from_model: get the noise rms given the noise model
    - set_image_to_mock: set the data to a mock realization given a model from Model()

    ToDo
    ----------
    - Add functions to degrade / reproject the data while dealing 
    with power spectrum properties
    - Add checks of the data format, consistency, type etc (e.g. header agree with data shape, 
    loaded covriance agrees with data shape, etc)

    """

    #==================================================
    # Initialize the data object
    #==================================================

    def __init__(self,
                 image,
                 header,
                 mask=None,
                 psf_fwhm=0*u.arcsec,
                 transfer_function=None,
                 noise_rms=None,
                 noise_covmat=None,
                 noise_model=None,
                 noise_mc=None,
                 silent=False,
                 output_dir='./pitszi_output',
    ):
        """
        Initialize the data object.
        
        Parameters
        ----------
        - image (2d np.array): data
        - header (str): header of the data
        - mask (2d np.array): mask associated with image
        - psf_fwhm (quantity): the FWHM of the PSF as a quantity homogeneous to arcsec
        - transfer_function (dict): the transfer function as {'k':k*u.arcsec, 'TF':TF}
        - noise_rms (2d np.array): rms associated with the image
        - noise_covmat (2d np.array): noise covariance matrix associated with flattened image
        - noise_model (list of 2 element): 
          b) returns the noise amplitude as a function of position. This can be a map, or a 
             function that depends on radius from the center in arcsec.
          a) function that give the noise power spectrum, once normalized by its amplitude,
             as a function of k in arcsec^-1
        - noise_mc (3d np array): Monte Carlo noise realizations
        - silent (bool): set to true to avoid printing information
        - output_dir (str): directory where saving outputs

        """

        if not silent:
            title.show_data()
        
        kref = np.linspace(0, 1, 1000)*u.arcsec**-1
        
        # Image data as a 2D np array, in Compton parameter unit
        self.image = image
        
        # Header of the image data
        self.header = header

        # Mask associated with the data
        if mask is None:
            self.mask = image * 0 + 1
        else:
            self.mask = mask

        # The PSF
        self.psf_fwhm = psf_fwhm

        # Transfer function, i.e. filtering as a function of k
        if transfer_function is None:
            self.transfer_function = {'k':kref,
                                      'TF':kref.value*0 + 1}
        else:
            self.transfer_function = transfer_function

        # Description of the noise
        self.noise_rms    = noise_rms
        self.noise_covmat = noise_covmat
        self.noise_model  = noise_model
        self.noise_mc     = noise_mc

        # Admin
        self.silent     = silent
        self.output_dir = output_dir

        
    #==================================================
    # Print parameters
    #==================================================
    
    def print_param(self):
        """
        Print the current parameters describing the data.
        
        Parameters
        ----------
        
        Outputs
        ----------
        The parameters are printed in the terminal
        
        """

        print('=====================================================')
        print('=============== Current Data() state ================')
        print('=====================================================')
        
        pp = pprint.PrettyPrinter(indent=4)
        
        par = self.__dict__
        keys = list(par.keys())
        
        for k in range(len(keys)):
            print(('--- '+(keys[k])))
            print(('    '+str(par[keys[k]])))
            print(('    '+str(type(par[keys[k]]))+''))
        print('=====================================================')


    #==================================================
    # Save parameters
    #==================================================
    
    def save_data(self):
        """
        Save the current data object.
        
        Parameters
        ----------
            
        Outputs
        ----------
        The parameters are saved in the output directory

        """

        # Create the output directory if needed
        if not os.path.exists(self.output_dir): os.mkdir(self.output_dir)

        # Save
        with open(self.output_dir+'/data_parameters.pkl', 'wb') as pfile:
            #pickle.dump(self.__dict__, pfile, pickle.HIGHEST_PROTOCOL)
            dill.dump(self.__dict__, pfile)

        # Text file for user
        par = self.__dict__
        keys = list(par.keys())
        with open(self.output_dir+'/data_parameters.txt', 'w') as txtfile:
            for k in range(len(keys)):
                txtfile.write('--- '+(keys[k])+'\n')
                txtfile.write('    '+str(par[keys[k]])+'\n')
                txtfile.write('    '+str(type(par[keys[k]]))+'\n')

                
    #==================================================
    # Load parameters
    #==================================================
    
    def load_data(self, param_file):
        """
        Read the a given parameter file to re-initialize the inference object.
        
        Parameters
        ----------
        param_file (str): the parameter file to be read
            
        Outputs
        ----------
            
        """

        with open(param_file, 'rb') as pfile:
            par = dill.load(pfile)
            
        self.__dict__ = par
        

    #==================================================
    # Set reference noise model
    #==================================================

    def set_nika2_reference_noise_model(self):
        """
        Set the noise model according to a typical baseline model.
        
        Parameters
        ----------
        
        """
        
        # output unitless
        self.noise_model[0] = lambda r_arcsec: 1 + np.exp((r_arcsec-200)/80)
        
        # output in [y] x arcsec^2
        self.noise_model[1] = lambda k_arcsec: 5e-9 + 15e-9 * (k_arcsec*60)**-1
        
        
    #==================================================
    # Set the Transfer function to a reference model
    #==================================================

    def set_nika2_reference_tf(self, k=np.linspace(0, 1, 1000)*u.arcsec**-1):
        """
        Set the transfer function to a parameteric NIKA2 baseline TF
        
        Parameters
        ----------
        - k (1d np quantity array ): the wavenumber
        
        """
        
        kfov = 1/(6*u.arcmin)
        TF = 1 - np.exp(-(k/kfov).to_value(''))
        transfer_function = {'k':k, 'TF':TF}
        self.transfer_function = transfer_function

        return transfer_function
    
    
    #==================================================
    # Compute noise MC realization from a noise model
    #==================================================
    
    def get_noise_monte_carlo_from_model(self,
                                         center=None,
                                         Nmc=1,
                                         seed=None):
        """
        Compute a noise MC realization given a noise Pk
        and a radial dependence
        
        Parameters
        ----------
        - center (SkyCoord): the reference center for the noise
        - Nmc (int): the number of monte carlo
        - seed (int): the numpy random seed for reproducible results
    
        Outputs
        ----------
        - noise (nd array): the noise monte carlo cube

        """

        # Check that the model was filled
        if type(self.noise_model) != list:
            if not self.silent:
                print('The noise model [noise_model_radial, noise_model_pk] should be a list')
                print('Cannot run get_noise_monte_carlo_from_model.')
            return

        #---------- Dependence in terms of k
        # Set a seed
        np.random.seed(seed)

        # Grid info
        header = self.header
        Nx = header['Naxis1']
        Ny = header['Naxis2']
        reso_arcsec = np.abs(header['CDELT2'])*3600
    
        # Define the wavevector
        k_x = np.fft.fftfreq(Nx, reso_arcsec) # 1/arcsec
        k_y = np.fft.fftfreq(Ny, reso_arcsec)
        k2d_x, k2d_y = np.meshgrid(k_x, k_y, indexing='xy')
        k2d_norm = np.sqrt(k2d_x**2 + k2d_y**2)
        k2d_norm_flat = k2d_norm.flatten()
    
        # Compute Pk noise
        P2d_k_grid = self.noise_model[1](k2d_norm)

        # Clean Pk
        wneg = P2d_k_grid < 0
        P2d_k_grid[wneg] = 0   # Negative Pk can happen for interpolation of poorly defined models
        if np.sum(wneg)>0 and not self.silent:
            print('----- The Pk noise model function implies some Pk<0. These points are set to 0.')
        P2d_k_grid[k2d_norm == 0] = 0

        # Generate the normalized noise
        amplitude =  np.sqrt(P2d_k_grid / reso_arcsec**2)
        noise = np.random.normal(0, 1, size=(Nmc, Ny, Nx))
        noise = np.fft.fftn(noise, axes=(1,2))
        noise = np.real(np.fft.ifftn(noise * amplitude, axes=(1,2)))
        
        #---------- Account for a radial dependence
        # Case of a function of radius
        if type(self.noise_model[0]) == type(lambda x: 1):
            ramap, decmap = map_tools.get_radec_map(header)
            if center is None:
                RA0 = np.mean(ramap)
                Dec0 = np.mean(decmap)
                center = SkyCoord(RA0*u.deg, Dec0*u.deg, frame="icrs")
            
            dist_map = map_tools.greatcircle(ramap, decmap,
                                             center.icrs.ra.to_value('deg'), center.icrs.dec.to_value('deg'))
            noise_radial_dep = self.noise_model[0](dist_map*3600)
            noise = noise * np.transpose(np.dstack([noise_radial_dep]*Nmc), axes=(2,0,1))

        # Case of a map
        if type(self.noise_model[0]) == type(self.image):
            if self.noise_model[0].shape != (Ny, Nx):
                raise ValueError('The radial dependence of the model (a map) does not match the data shape')
            noise = noise * np.transpose(np.dstack([self.noise_model[0]]*Nmc), axes=(2,0,1))

        # Clean noise in case
        wbad = np.isnan(noise) * np.isinf(noise)
        noise[wbad] = np.nan

        #---------- Remove first axis if Nmc is one
        if Nmc == 1:
            noise = noise[0]
        
        return noise


    #==================================================
    # Compute noise MC realization from the covariance
    #==================================================
    
    def get_noise_monte_carlo_from_covariance(self,
                                              Nmc=1,
                                              seed=None):
        """
        Compute a noise MC realization from the noise covariance
        matrix.
        
        Parameters
        ----------
        - Nmc (int): the number of map to compute in the cube
        - seed (int): the 
        
        Outputs
        ----------
        - noise (nd array): noise MC cube
        """

        # Check that the covariance exists
        if self.noise_covmat is None:
            raise ValueError('The noise covariance matrix is undefined, but needed here')

        # Set a seed
        np.random.seed(seed)

        # Get the flat noise realization
        target_shape = self.image.shape
        noise_mc_flat = np.random.multivariate_normal(np.zeros(self.noise_covmat.shape[0]),
                                                      self.noise_covmat,
                                                      size=Nmc)
        # Reshape it to maps
        if Nmc == 1:
            noise_mc = noise_mc_flat.reshape(target_shape[0], target_shape[1])
        else:
            noise_mc = noise_mc_flat.reshape(Nmc, target_shape[0], target_shape[1])

        return noise_mc

    
    #==================================================
    # Get the noise covariance from model MC realizations
    #==================================================

    def get_noise_covariance_from_model(self,
                                        center=None,
                                        Nmc=1000,
                                        seed=None):
        """
        Get the noise covariance from the noise model
        
        Parameters
        ----------
        - center (SkyCoord): the reference center for the noise
        - Nmc (int): the number of monte carlo
        - seed (int): the numpy random seed for reproducible results

        Outputs
        ----------
        - covmat (nd array): the noise covariance matrix

        """

        if not self.silent:
            print('Start computing the covariance matrix from MC simulations')
            print('This can take significant time')
            
        noise_mc = self.get_noise_monte_carlo_from_model(center=center,
                                                         Nmc=Nmc,
                                                         seed=seed)
        covmat = 0
        for i in range(Nmc):
            mflat = noise_mc[i,:,:].flatten()
            covmat += np.matmul(mflat[:,None], mflat[None,:])
        covmat /= Nmc

        return covmat


    #==================================================
    # Get the noise covariance from user MC realizations
    #==================================================

    def get_noise_covariance_from_noise_mc(self):
        """
        Get the noise covariance from the noise MC realizations
        
        Parameters
        ----------

        Outputs
        ----------
        - covmat (nd array): the noise covariance matrix

        """

        if self.noise_mc is None:
            raise ValueError('The noise_mc was not definied, but it is required for getting the covariance here')
        
        if not self.silent:
            print('Start computing the covariance matrix from MC simulations')
            print('This can take significant time')
            
        noise_mc = self.noise_mc
        Nmc = noise_mc.shape[0]
        
        covmat = 0
        for i in range(Nmc):
            mflat = noise_mc[i,:,:].flatten()
            covmat += np.matmul(mflat[:,None], mflat[None,:])
        covmat /= Nmc

        return covmat


    #==================================================
    # Save the noise covariance matrix
    #==================================================
    
    def save_noise_covariance(self, filename=None):
        """
        Save the noise covariance matrix to avoid long
        computation time
        
        Parameters
        ----------
        - filename (str): full path to the covariance matrix file
        
        Outputs
        ----------
        - file is saved
        """

        if filename is None:
            myfilename = self.output_dir+'/data_image_noise_covariance_matrix.pkl'
        else:
            myfilename = filename
        
        with open(myfilename, 'wb') as pfile:
            pickle.dump(self.noise_covmat, pfile, pickle.HIGHEST_PROTOCOL)


    #==================================================
    # Read saved noise covariance matrix
    #==================================================
    
    def load_noise_covariance(self, filename=None):
        """
        Read the noise covariance matrix
        
        Parameters
        ----------
        - filename (str): full path to the covariance matrix file
        
        Outputs
        ----------
        
        """

        if filename is None:
            myfilename = self.output_dir+'/data_image_noise_covariance_matrix.pkl'
        else:
            myfilename = filename
            
        with open(myfilename, 'rb') as pfile:
            cov = pickle.load(pfile)
            
        self.noise_covmat = cov
    

    #==================================================
    # Measure noise from JackKnife
    #==================================================
    
    def set_noise_model_from_jackknife(self,
                                       jkmap,
                                       normmap,
                                       reso,
                                       Nbin=20,
                                       scale='log',
                                       show=False):
        """
        Useful function to derive the noise model from a jacknife by
        interpolating the power spectrum of the homogeneized image.
        The validity of the output noise model need to be checked though.
        In particular, this assumes the power spectrum to be the same 
        over the whole map once normalization is applied. Unwanted 
        interpolation effect may also appear.
        
        Parameters
        ----------
        - jkmap (np array): a jack knife map
        - normmap (np array): the normalization map. jkmap/normmap should have homogeneous noise
        - reso (float): the map resolution in arcsec
        - Nbin (int): number of bin for the Pk
        - scale (str): log or lin, to define the scale used in Pk
        - show (bool): check what is done with plots

        Outputs
        ----------
        The noise model is set from the jackknife
        """
        
        # General info
        Nx, Ny = jkmap.shape
        
        w1 = (normmap > 0) * ~np.isnan(normmap) * ~np.isinf(normmap)
        w2 = (jkmap != 0) * ~np.isnan(jkmap) * ~np.isinf(jkmap)
        fsky = np.sum((w1*w2)) / Nx / Ny
        if not self.silent:
            if np.sum(~w1) > 0 or np.sum(~w2) > 0:
                print('WARNING: some pixels are bad (fsky='+str(fsky)+'). This may affect the recovered noise model')
                
        # Extract the normalized map
        img = jkmap/normmap
        img[~(w1*w2)] = 0
        
        # Extract and fit the Pk
        k, pk = utils_pk.extract_pk2d(img, reso, Nbin=Nbin, scalebin=scale)
        pk = pk / fsky
        w = ~np.isnan(pk)
        #spec_mod = interp1d(k[w], pk[w], bounds_error=False, fill_value="extrapolate", kind='linear')
        spec_mod = interp1d(np.log10(k[w]), np.log10(pk[w]), bounds_error=False, fill_value="extrapolate", kind='linear')

        # Show
        if show:
            import matplotlib.pyplot as plt

            plt.figure(1)
            plt.imshow(jkmap/normmap)
            plt.colorbar()
            plt.title('JK / norm')

            plt.figure(2)
            plt.hist((jkmap/normmap).flatten(), bins=50)
            plt.yscale('log')
            plt.xlabel('Pixel value')
            plt.ylabel('PDF')
            print('JK/norm standard deviation: ', np.nanstd(jkmap/normmap))
            
            plt.figure(3)
            plt.loglog(k, pk, color='k', marker='o', label='Measured Pk')
            my_k = np.logspace(np.log10(np.amin(k)), np.log10(np.amax(k)), 50)
            plt.loglog(my_k, 10**spec_mod(np.log10(my_k)), color='r', label='Interpolation in 50 log spaced k')
            plt.xlabel('$k$ (1/arcsec)')
            plt.ylabel('$P_k$ (arcsec$^2$)')
            plt.legend()
            plt.show()

        # Fix the model
        #self.noise_model = [normmap, lambda k_arcsec: spec_mod(k_arcsec)]
        self.noise_model = [normmap, lambda k_arcsec: 10**spec_mod(np.log10(k_arcsec))]
        

    #==================================================
    # Set noise model from MC
    #==================================================
    
    def set_noise_model_from_mc(self,
                                Nbin=30,
                                scale='log'):
        """
        Derive the noise model by interpolating the noise PK from MC realizations
        after homogeneizing the noise        
        
        Parameters
        ----------
        - Nbin (int): number of bin for the Pk
        - scale (str): log or lin, to define the scale used in Pk
        
        Outputs
        ----------
        The noise model is set from the MC
        """

        #----- Sanity check
        if self.noise_mc is None:
            raise ValueError('The noise_mc should be defined for this function')
        
        #----- General info
        Nmc, Ny, Nx = self.noise_mc.shape
        if Nmc < 2: raise ValueError('The number of MC should be large, at least larger than 1.')
        reso = np.abs(self.header['CDELT1'])*3600
        normmap = np.std(self.noise_mc, axis=0)
        
        w = (normmap > 0) * ~np.isnan(normmap) * ~np.isinf(normmap)
        if not self.silent:
            if np.sum(~w) > 0:
                print('WARNING: some pixels are bad. This may affect the recovered noise model')
        
        #----- Extract the pk for each normalized MC
        pk_mc = np.zeros((Nmc,Nbin))
        for i in range(Nmc):
            img = self.noise_mc[i,:,:] / normmap
            img[~w] = 0
        
            k, pk_i = utils_pk.extract_pk2d(img, reso, Nbin=Nbin, scalebin=scale)
            pk_mc[i, :] += pk_i

        pk = np.mean(pk_mc, axis=0)
            
        w = ~np.isnan(pk)
        spec_mod = interp1d(k[w], pk[w], bounds_error=False, fill_value="extrapolate", kind='linear')
        
        #----- Fix the model
        self.noise_model = [normmap, lambda k_arcsec: spec_mod(k_arcsec)]
        


    #==================================================
    # Set noise model from MC
    #==================================================
    
    def set_noise_model_from_covariance(self,
                                        Nmc=1000,
                                        Nbin=30,
                                        scale='log'):
        """
        Derive the noise model by interpolating the noise PK from MC realizations
        after homogeneizing the noise, derive from the covariance matrix
        
        Parameters
        ----------
        - Nmc (int): number of MC to do
        - Nbin (int): number of bin for the Pk
        - scale (str): log or lin, to define the scale used in Pk
        
        Outputs
        ----------
        The noise model is set from the MC
        """

        #----- Sanity check
        if self.noise_covmat is None:
            raise ValueError('The noise_covmat should be defined for this function')

        #----- Get the noise MC
        noise_mc = self.get_noise_monte_carlo_from_covariance(Nmc=Nmc)
        
        #----- General info
        Nmc, Ny, Nx = noise_mc.shape
        if Nmc < 2: raise ValueError('The number of MC should be large, at least larger than 1.')
        reso = np.abs(self.header['CDELT1'])*3600
        normmap = np.std(noise_mc, axis=0)
        
        w = (normmap > 0) * ~np.isnan(normmap) * ~np.isinf(normmap)
        if not self.silent:
            if np.sum(~w) > 0:
                print('WARNING: some pixels are bad. This may affect the recovered noise model')
        
        #----- Extract the pk for each normalized MC
        pk_mc = np.zeros((Nmc,Nbin))
        for i in range(Nmc):
            img = noise_mc[i,:,:] / normmap
            img[~w] = 0
        
            k, pk_i = utils_pk.extract_pk2d(img, reso, Nbin=Nbin, scalebin=scale)
            pk_mc[i, :] += pk_i

        pk = np.mean(pk_mc, axis=0)
            
        w = ~np.isnan(pk)
        spec_mod = interp1d(k[w], pk[w], bounds_error=False, fill_value="extrapolate", kind='linear')
        
        #----- Fix the model
        self.noise_model = [normmap, lambda k_arcsec: spec_mod(k_arcsec)]
        
        
    #==================================================
    # Get the noise rms given a noise MC
    #==================================================
    
    def get_noise_rms_from_mc(self,
                              smooth_fwhm=0*u.arcsec):
        """
        Compute the noise rms from given noise MC
        
        Parameters
        ----------
        - smooth_fwhm (quantity): the gaussian FWHM to smooth the MC prior rms 

        Outputs
        ----------
        - rms (np array): The noise rms
        """
        
        #----- First get noise MC
        noise_mc = copy.copy(self.noise_mc)

        #----- Have the possibility to smooth the noise MC
        if smooth_fwhm>0:
            sigma2fwhm = 2 * np.sqrt(2*np.log(2))
            reso = np.abs(self.header['CDELT1'])
            sigma = smooth_fwhm.to_value('deg')/sigma2fwhm/reso
            noise_mc = gaussian_filter(noise_mc, sigma=[0,sigma,sigma])
        
        #----- Compute the rms
        rms = np.std(noise_mc, axis=0)
        
        #----- Get the rms
        return rms
        

    #==================================================
    # Get the noise rms given a noise MC
    #==================================================
    
    def get_noise_rms_from_covariance(self,
                                 Nmc=1000,
                                 smooth_fwhm=0*u.arcsec,
                                 noise_seed=None):
        """
        Compute the noise rms from the covariance matrix
        
        Parameters
        ----------
        - Nmc (int): number of MC to do
        - smooth_fwhm (quantity): the gaussian FWHM to smooth the MC prior rms 
        - noise_seed (int): the numpy random seed for reproducible results

        Outputs
        ----------
        - rms (np array): The noise rms
        """
        
        #----- First get noise MC
        noise_mc = self.get_noise_monte_carlo_from_covariance(Nmc=Nmc, seed=noise_seed)

        #----- Have the possibility to smooth the noise MC
        if smooth_fwhm>0:
            sigma2fwhm = 2 * np.sqrt(2*np.log(2))
            reso = np.abs(self.header['CDELT1'])
            sigma = smooth_fwhm.to_value('deg')/sigma2fwhm/reso
            noise_mc = gaussian_filter(noise_mc, sigma=[0,sigma,sigma])
        
        #----- Compute the rms
        rms = np.std(noise_mc, axis=0)
        
        #----- Get the rms
        return rms
    

    #==================================================
    # Get the noise rms given a model
    #==================================================
    
    def get_noise_rms_from_model(self,
                                 Nmc=1000,
                                 smooth_fwhm=0*u.arcsec,
                                 center=None,
                                 noise_seed=None):
        """
        Compute the noise rms from the noise model
        
        Parameters
        ----------
        - Nmc (int): number of MC to do
        - smooth_fwhm (quantity): the gaussian FWHM to smooth the MC prior rms 
        - center (SkyCoord): the reference center for the noise
        - noise_seed (int): the numpy random seed for reproducible results
        
        Outputs
        ----------
        - rms (np array): The noise rms
        """
        
        #----- First get noise MC
        noise_mc = self.get_noise_monte_carlo_from_model(center=center,
                                                         Nmc=Nmc,
                                                         seed=noise_seed)

        #----- Have the possibility to smooth the noise MC
        if smooth_fwhm>0:
            sigma2fwhm = 2 * np.sqrt(2*np.log(2))
            reso = np.abs(self.header['CDELT1'])
            sigma = smooth_fwhm.to_value('deg')/sigma2fwhm/reso
            noise_mc = gaussian_filter(noise_mc, sigma=[0,sigma,sigma])
        
        #----- Compute the rms
        rms = np.std(noise_mc, axis=0)
        
        #----- Get the rms
        return rms

    
    #==================================================
    # Generate Mock data
    #==================================================

    def set_image_to_mock(self,
                          model_input,
                          model_seed=None,
                          model_no_fluctuations=False,
                          use_model_header=False,
                          noise_origin='model',
                          noise_center=None,
                          noise_seed=None):
        """
        Set the data image to a mock realization given a model
        
        Parameters
        ----------
        - model (class Model object): the model
        - model_seed (bool): set to a number for reproducible fluctuations
        - model_no_fluctuations (bool): set to true when the pure spherical model is requested
        - use_model_header (bool): set to true to replacde data header 
        with the current model header. Otherwise the model header is 
        set to the one of the data.
        - noise_origin (str): can be 'covariance', 'model', 'MC', or 'none'
        - noise_center (SkyCoord): the reference center for the noise
        - model_seed (bool): set to a number for reproducible noise

        """
        
        # copy the model so that the input is not modified
        model = copy.deepcopy(model_input)

        # Match the header as requested
        if use_model_header:
            self.header = model.get_map_header()
            if not self.silent:
                msg1 = 'WARNING: the data header is fixed to the model header.'
                msg2 = '         This may affect consistency with other data properties (mask, noise, etc)'
                print(msg1)
                print(msg2)
        else:
            model.map_header = self.header
            
        # Get the raw image model
        input_model = model.get_sz_map(seed=model_seed, no_fluctuations=model_no_fluctuations)
        
        # Convolve with instrument response function
        convolved_model = utils_pk.apply_transfer_function(input_model,
                                                           model.get_map_reso().to_value('arcsec'),
                                                           self.psf_fwhm.to_value('arcsec'),
                                                           self.transfer_function,
                                                           apps_TF_LS=True, apps_beam=True)
        
        # Add noise realization
        if noise_origin == 'model':
            noise_mc = self.get_noise_monte_carlo_from_model(center=noise_center, Nmc=1, seed=noise_seed)
        elif noise_origin == 'covariance':
            noise_mc = self.get_noise_monte_carlo_from_covariance(Nmc=1, seed=noise_seed)
        elif noise_origin == 'MC':
            if len(self.noise_mc.shape) == 3:
                noise_mc = self.noise_mc[0,:,:]
            elif len(self.noise_mc) == 2:
                noise_mc = self.noise_mc
            else:
                raise ValueError('The noise_mc is not correct')
        elif noise_origin == 'none':
            noise_mc = 0
        else:
            raise ValueError('noise_origin should be "model", "covariance", or "none"')
        image_mock = convolved_model + noise_mc

        # Set the image and the mask if not correct anymore
        self.image = image_mock     

        return image_mock
        
